"""Prepare user data for weight generation."""
import time
from abc import ABC
from abc import abstractmethod
from pathlib import Path
from typing import Any
from typing import List
from typing import Union

import geopandas as gpd
import numpy as np
import pandas as pd
import rioxarray as rxr
import xarray as xr
from pyproj.crs import CRS

from gdptools.data.agg_gen_data import AggData
from gdptools.data.odap_cat_data import CatGrids
from gdptools.data.odap_cat_data import CatParams
from gdptools.data.weight_gen_data import WeightData
from gdptools.helpers import build_subset
from gdptools.helpers import build_subset_tiff
from gdptools.utils import _check_for_intersection
from gdptools.utils import _check_for_intersection_nc
from gdptools.utils import _get_cells_poly
from gdptools.utils import _get_data_via_catalog
from gdptools.utils import _get_shp_bounds_w_buffer
from gdptools.utils import _get_shp_file
from gdptools.utils import _get_top_to_bottom
from gdptools.utils import _read_shp_file


class UserData(ABC):
    """Prepare data for different sources for weight generation."""

    @abstractmethod
    def __init__(self) -> None:
        """Init class."""
        pass

    @abstractmethod
    def prep_wght_data(self) -> WeightData:
        """Abstract interface for generating weight data."""
        pass

    @abstractmethod
    def prep_agg_data(self) -> AggData:
        """Abstract method for preparing data for aggregation."""
        pass

    @abstractmethod
    def get_vars(self) -> list[str]:
        """Return a list of variables."""
        pass

    @abstractmethod
    def get_feature_id(self) -> str:
        """Abstract method for returning the id_feature parameter."""
        pass


class ODAPCatData(UserData):
    """Instance of UserData using OPeNDAP catalog data."""

    def __init__(
        self,
        *,
        param_dict: dict[str, dict],
        grid_dict: dict[str, dict],
        f_feature: Union[str, Path, gpd.GeoDataFrame],
        id_feature: str,
        period: List[str],
    ) -> None:
        """Initialize PrepODAPCatData class.

        Args:
            param_dict (dict[str, dict]): Parameter metadata from OPeNDAP catalog.
            grid_dict (dict[str, dict]): Grid metadata from OPeNDAP catalog.
            f_feature (Union[str, Path, gpd.GeoDataFrame]): GeoDataFrame or any
                path-like object that can be read by geopandas.read_file().
            id_feature (str): Header in GeoDataFrame to use as index for weights.
            period (List[str]): List containing date strings defining start and end
                time slice for aggregation.
        """
        self.param_dict = param_dict
        self.grid_dict = grid_dict
        self.f_feature = f_feature
        self.id_feature = id_feature
        self.period = pd.to_datetime(period)
        self._heck_input_dicts()
        self.gdf = _read_shp_file(self.f_feature)

    def get_feature_id(self) -> str:
        """Return id_feature."""
        return self.id_feature

    def get_vars(self) -> list[str]:
        """Return list of param_dict keys, proxy for varnames."""
        return list(self.param_dict.keys())

    def prep_agg_data(self, key: str) -> AggData:
        """Prepare ODAPCatData data for aggregation methods.

        Args:
            key (str): _description_

        Returns:
            AggData: _description_
        """
        cp = self._create_catparam(key=key)
        cg = self._create_catgrid(key=key)

        is_intersect, is_degrees, is_0_360 = _check_for_intersection(
            cat_params=cp, cat_grid=cg, gdf=self.gdf
        )

        gdf, gdf_bounds = _get_shp_file(
            shp_file=self.gdf, cat_grid=cg, is_degrees=is_degrees
        )

        rotate_ds = bool((not is_intersect) & is_degrees & (not is_0_360))
        ds_ss = _get_data_via_catalog(
            cat_params=cp,
            cat_grid=cg,
            bounds=gdf_bounds,
            begin_date=self.period[0],
            end_date=self.period[1],
            rotate_lon=rotate_ds,
        )
        return AggData(
            variable=key,
            cat_param=cp,
            cat_grid=cg,
            da=ds_ss,
            feature=gdf,
            id_feature=self.id_feature,
            period=self.period,
        )

    def prep_wght_data(self) -> WeightData:
        """Prepare and return WeightData for weight generation."""
        # For weight generation we just need param and grid dict from one variable
        # passed to CatParms and CatGrids.
        self._heck_input_dicts()
        keys = self.get_vars()
        cp = self._create_catparam(key=keys[0])
        cg = self._create_catgrid(key=keys[0])
        gdf = _read_shp_file(self.f_feature)
        self._check_id_feature(gdf)
        is_intersect, is_degrees, is_0_360 = _check_for_intersection(
            cat_params=cp, cat_grid=cg, gdf=gdf
        )

        gdf, gdf_bounds = _get_shp_file(
            shp_file=gdf, cat_grid=cg, is_degrees=is_degrees
        )

        rotate_ds = bool((not is_intersect) & is_degrees & (not is_0_360))
        ds_ss = _get_data_via_catalog(
            cat_params=cp,
            cat_grid=cg,
            bounds=gdf_bounds,
            begin_date=self.period[0],
            rotate_lon=rotate_ds,
        )

        # tsrt = time.perf_counter()
        # ds_ss.load()
        # tend = time.perf_counter()
        # print(f"loading {cp.varname} in {tend-tsrt:0.4f} seconds")

        tsrt = time.perf_counter()
        gdf_grid = _get_cells_poly(
            xr_a=ds_ss,
            x=cg.X_name,
            y=cg.Y_name,
            crs_in=cg.proj,
        )
        tend = time.perf_counter()
        print(f"grid cells generated in {tend-tsrt:0.4f} seconds")

        return WeightData(feature=gdf, id_feature=self.id_feature, grid_cells=gdf_grid)

    def _check_id_feature(self, gdf):
        """Check id_feature in gdf."""
        if self.id_feature not in gdf.columns[:]:
            raise ValueError(
                f"shp_poly_idx: {self.id_feature} not in gdf columns: {gdf.columns}"
            )

    def _create_catgrid(self, key: str):
        """Create CatGrids."""
        return CatGrids(**self.grid_dict.get(key))

    def _create_catparam(self, key: str):
        """Create CatParams."""
        return CatParams(**self.param_dict.get(key))

    def _heck_input_dicts(self):
        """Check input dicts are of equal length and with at least one entry."""
        if len(self.param_dict) != len(self.grid_dict):
            raise ValueError(
                "Mismatch in key,value pairs. param_dict and grid_dict \
                should have tehe same number of keys."
            )
        if len(self.param_dict) < 1:
            raise ValueError("param_dict should have at least 1 key,value pair")


class UserCatData(UserData):
    """Instance of UserData using minium input variables to map to ODAPCatData."""

    def __init__(
        self,
        ds: Union[str, Path, xr.Dataset],
        proj_ds: Any,
        x_coord: str,
        y_coord: str,
        t_coord: str,
        var: List[str],
        f_feature: Union[str, Path, gpd.GeoDataFrame],
        proj_feature: str,
        id_feature: str,
        period: List[str],
    ) -> None:
        """Contains data preperation methods based on UserData.

        Args:
            ds (Union[str, Path, xr.Dataset]): _description_
            proj_ds (Any): _description_
            x_coord (str): _description_
            y_coord (str): _description_
            t_coord (str): _description_
            var (List[str]): _description_
            f_feature (Union[str, Path, gpd.GeoDataFrame]): _description_
            proj_feature (str): _description_
            id_feature (str): _description_
            period (List[str]): List of 2 datetime strings representing start and end of
                period of interest.

        Raises:
            TypeError: _description_
        """
        self.ds = ds
        self.proj_ds = proj_ds
        self.x_coord = x_coord
        self.y_coord = y_coord
        self.t_coord = t_coord
        # self.period = pd.to_datetime(period)
        self.period = period
        if not isinstance(var, List):
            raise TypeError(f"Input var {var} should be a list of variables")
        self.var = var
        self.f_feature = f_feature
        self.proj_feature = proj_feature
        self.id_feature = id_feature

    @classmethod
    def __repr__(cls) -> str:
        """Print class name."""
        return f"Class is {cls.__name__}"

    def get_feature_id(self) -> str:
        """Return id_feature."""
        return self.id_feature

    def get_vars(self) -> list[str]:
        """Return list of vars in data."""
        return self.var

    def prep_agg_data(self, key: str) -> AggData:
        """Prep AggData from UserData."""
        ds = self._get_xr_dataset()

        gdf_in = self._get_geodataframe()

        gdf_in.set_crs(self.proj_feature, inplace=True)

        gdf_bounds = _get_shp_bounds_w_buffer(
            gdf=gdf_in, ds=ds, crs=self.proj_ds, lon=self.x_coord, lat=self.y_coord
        )

        is_intersect, is_degrees, is_0_360 = _check_for_intersection_nc(
            ds=ds,
            x_name=self.x_coord,
            y_name=self.y_coord,
            proj=self.proj_ds,
            gdf=gdf_in,
        )

        if bool((not is_intersect) & is_degrees & (not is_0_360)):  # rotate
            ds.coords[self.x_coord] = (ds.coords[self.x_coord] + 180) % 360 - 180
            ds = ds.sortby(ds[self.x_coord])

        # calculate toptobottom (order of latitude coords)
        ttb = _get_top_to_bottom(ds, self.y_coord)

        subset_dict = build_subset(
            bounds=gdf_bounds,
            xname=self.x_coord,
            yname=self.y_coord,
            tname=self.t_coord,
            toptobottom=ttb,
            date_min=self.period[0],
            date_max=self.period[1],
        )

        data_ss = ds[key].sel(**subset_dict)
        cplist = self._get_user_catparams(key=key, da=data_ss)
        cglist = self._get_user_catgrids(key=key, da=data_ss)

        param_dict = dict(zip([key], cplist))  # noqa B905
        grid_dict = dict(zip([key], cglist))  # noqa B905

        # print(param_dict)

        return AggData(
            variable=key,
            cat_param=CatParams(**param_dict.get(key)),
            cat_grid=CatGrids(**grid_dict.get(key)),
            da=data_ss,
            feature=gdf_in,
            id_feature=self.id_feature,
            period=self.period,
        )

    def _get_user_catgrids(self, key: str, da: xr.DataArray):
        """Get CatGrids from UserData."""
        cglist = []
        cg = CatGrids(
            X_name=self.x_coord,
            Y_name=self.y_coord,
            proj=self.proj_ds,
            resX=max(np.diff(da[self.x_coord].values)),
            resY=max(np.diff(da[self.y_coord].values)),
            toptobottom=_get_top_to_bottom(da, self.y_coord),
        )
        cglist.append(cg.dict())
        return cglist

    def _get_ds_var_attrs(self, da: xr.DataArray, attr: str) -> str:
        """Return var attributes.

        Args:
            da (xr.DataArray): _description_
            attr (str): _description_

        Returns:
            str: _description_
        """
        try:
            return da.attrs.get(attr)
        except KeyError:
            return "None"

    def _get_user_catparams(self, key: str, da: xr.DataArray):
        """Get CatParams from UserData."""
        cplist = []
        cp = CatParams(
            URL="",
            varname=key,
            long_name=self._get_ds_var_attrs(da, "long_name"),
            T_name=self.t_coord,
            units=self._get_ds_var_attrs(da, "units"),
        )
        cplist.append(cp.dict())
        return cplist

    def prep_wght_data(self) -> WeightData:
        """Prepare and return WeightData for weight generation."""
        ds = self._get_xr_dataset()

        gdf_in = self._get_geodataframe()

        gdf_in.set_crs(self.proj_feature, inplace=True)

        gdf_bounds = _get_shp_bounds_w_buffer(
            gdf=gdf_in, ds=ds, crs=self.proj_ds, lon=self.x_coord, lat=self.y_coord
        )

        is_intersect, is_degrees, is_0_360 = _check_for_intersection_nc(
            ds=ds,
            x_name=self.x_coord,
            y_name=self.y_coord,
            proj=self.proj_ds,
            gdf=gdf_in,
        )

        if bool((not is_intersect) & is_degrees & (not is_0_360)):  # rotate
            print("rotating")
            ds.coords[self.x_coord] = (ds.coords[self.x_coord] + 180) % 360 - 180
            ds = ds.sortby(ds[self.x_coord])

        # calculate toptobottom (order of latitude coords)
        ttb = _get_top_to_bottom(ds, self.y_coord)

        subset_dict = build_subset(
            bounds=gdf_bounds,
            xname=self.x_coord,
            yname=self.y_coord,
            tname=self.t_coord,
            toptobottom=ttb,
            date_min=self.period[0],
        )

        # print(subset_dict)

        data_ss_wght = ds.sel(**subset_dict)

        start = time.perf_counter()
        grid_poly = _get_cells_poly(
            xr_a=data_ss_wght,
            x=self.x_coord,
            y=self.y_coord,
            crs_in=self.proj_ds,
        )
        end = time.perf_counter()
        print(
            f"Generating grid-cell bounds finished in {round(end-start, 2)} second(s)"
        )
        return WeightData(
            feature=gdf_in, id_feature=self.id_feature, grid_cells=grid_poly
        )

    def _get_geodataframe(self):
        """Get GeoDataFrame."""
        if not isinstance(self.f_feature, gpd.GeoDataFrame):
            return gpd.read_file(self.f_feature)
        else:
            return self.f_feature

    def _get_xr_dataset(self):
        """Get xarray.Dataset."""
        if not isinstance(self.ds, xr.Dataset):
            ds = xr.open_dataset(self.ds)
        else:
            ds = self.ds
        return ds


class UserTiffData(UserData):
    """Instance of UserData for zonal stats processing of geotiffs."""

    def __init__(
        self,
        var: str,
        ds: Union[str, xr.Dataset],
        proj_ds: Any,
        x_coord: str,
        y_coord: str,
        bname: str,
        band: int,
        f_feature: Union[str, Path, gpd.GeoDataFrame],
        id_feature: str,
        proj_feature: Any,
    ) -> None:
        """__init__ Initialize UserTiffData.

        UserTiffData is a structure that aids calculating zonal stats.

        Args:
            var (str): _description_
            ds (Union[str, xr.Dataset]): _description_
            proj_ds (Any): _description_
            x_coord (str): _description_
            y_coord (str): _description_
            bname (str): _description_
            band (int): _description_
            f_feature (Union[str, Path, gpd.GeoDataFrame]): _description_
            id_feature (str): _description_
            proj_feature (Any): _description_
        """
        self.varname = var
        self.x_coord = x_coord
        self.y_coord = y_coord
        self.bname = bname
        self.band = band
        self.toptobottom = None
        self._get_rxr_dataset(ds)
        self.proj_ds = proj_ds
        self._get_geodataframe(f_feature)
        self.id_feature = id_feature
        self.proj_feature = proj_feature
        self.f_feature = self.f_feature.sort_values(self.id_feature).dissolve(
            by=self.id_feature, as_index=False
        )
        self.f_feature.reset_index()

        self._check_xname()
        self._check_yname()
        self._check_band()
        self._check_crs()
        self._get_toptobottom()

    def get_vars(self) -> list[str]:
        """Return list of varnames."""
        return [self.varname]

    def get_feature_id(self) -> str:
        """Get Feature id."""
        return self.id_feature

    def prep_wght_data(self) -> WeightData:
        """Prepare data for weight generation."""
        pass

    def prep_agg_data(self, key: str) -> AggData:
        """Prepare data for aggregation or zonal stats."""
        bb_feature = _get_shp_bounds_w_buffer(
            gdf=self.f_feature,
            ds=self.ds,
            crs=self.proj_ds,
            lon=self.x_coord,
            lat=self.y_coord,
        )

        subset_dict = build_subset_tiff(
            bounds=bb_feature,
            xname=self.x_coord,
            yname=self.y_coord,
            toptobottom=self.toptobottom,
            bname=self.bname,
            band=self.band,
        )

        data_ss = self.ds.sel(**subset_dict)  # .rio.interpolate_na()
        # tstrt = time.perf_counter()
        # data_ss = data_ss.rio.interpolate_na(method='nearest')
        # tend = time.perf_counter()
        # print(f"fill missing data using nearest method in {tend - tstrt:0.4f} seconds")
        cplist = self._get_user_catparams(da=data_ss)
        cglist = self._get_user_catgrids(da=data_ss)
        param_dict = dict(zip([key], [cplist]))  # noqa B905
        grid_dict = dict(zip([key], [cglist]))  # noqa B905

        return AggData(
            variable=key,
            cat_param=CatParams(**param_dict.get(key)),
            cat_grid=CatGrids(**grid_dict.get(key)),
            da=data_ss,
            feature=self.f_feature,
            id_feature=self.id_feature,
            period=["None", "None"],
        )

    def _get_toptobottom(self):
        """Generate toptobottom."""
        yy = self.ds.coords[self.y_coord].values
        self.toptobottom = yy[0] <= yy[-1]

    def _check_xname(self):
        """Validate xname."""
        if self.x_coord not in self.ds.coords:
            raise ValueError(f"xname:{self.x_coord} not in {self.ds.coords}")

    def _check_yname(self):
        """Validate yname."""
        if self.y_coord not in self.ds.coords:
            raise ValueError(f"yname:{self.y_coord} not in {self.ds.coords}")

    def _check_band(self):
        """Validate band name."""
        if self.bname not in self.ds.coords:
            raise ValueError(f"band:{self.bname} not in {self.ds.coords}")

    def _check_crs(self):
        """Validate crs."""
        crs = CRS.from_user_input(self.proj_ds)
        if not isinstance(crs, CRS):
            raise ValueError(f"ds_proj:{self.proj_ds} not in valid")
        crs2 = CRS.from_user_input(self.proj_feature)
        if not isinstance(crs2, CRS):
            raise ValueError(f"ds_proj:{self.proj_feature} not in valid")

    def _get_rxr_dataset(self, ds: Union[str, xr.DataArray, xr.Dataset]):
        """Get xarray.Dataset."""
        if not isinstance(ds, (xr.Dataset, xr.DataArray)):
            self.ds = rxr.open_rasterio(ds)
        else:
            self.ds = ds

    def _get_geodataframe(self, f_feature: gpd.GeoDataFrame):
        """Get GeoDataFrame."""
        if not isinstance(f_feature, gpd.GeoDataFrame):
            self.f_feature = gpd.read_file(f_feature)
        else:
            self.f_feature = f_feature

    def _get_user_catgrids(self, da: xr.DataArray):
        """Get CatGrids from UserData."""
        cg = CatGrids(
            X_name=self.x_coord,
            Y_name=self.y_coord,
            proj=self.proj_ds,
            resX=max(np.diff(da[self.x_coord].values)),
            resY=max(np.diff(da[self.y_coord].values)),
            toptobottom=self.toptobottom,
        )
        return cg.dict()

    def _get_user_catparams(self, da: xr.DataArray):
        """Get CatParams from UserData."""
        cp = CatParams(
            URL="",
            varname=self.varname,
            long_name=self._get_ds_var_attrs(da, "long_name"),
            T_name="",
            units=self._get_ds_var_attrs(da, "units"),
        )
        return cp.dict()

    def _get_ds_var_attrs(self, da: xr.DataArray, attr: str) -> str:
        """Return var attributes.

        Args:
            da (xr.DataArray): _description_
            attr (str): _description_

        Returns:
            str: _description_
        """
        try:
            return da.attrs.get(attr)
        except KeyError:
            return "na"



"""Tests for .helper functions."""
from tempfile import NamedTemporaryFile
from tempfile import TemporaryDirectory

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
import xarray as xr

from gdptools import AggGen
from gdptools import UserCatData
from gdptools import WeightGen


@pytest.fixture()
def get_gdf() -> gpd.GeoDataFrame:
    """Create GeoDataFrame."""
    return gpd.read_file("./tests/data/DRB/DRB_4326.shp")


@pytest.fixture()
def get_xarray() -> xr.Dataset:
    """Create xarray Dataset."""
    return xr.open_dataset("./tests/data/DRB/o_of_b_test.nc")


@pytest.fixture()
def get_file_path(tmp_path):
    """Get temp file path."""
    return tmp_path / "test.csv"


@pytest.fixture()
def get_out_path(tmp_path):
    """Get temp file output path."""
    return tmp_path


data_crs = 4326
x_coord = "lon"
y_coord = "lat"
t_coord = "time"
sdate = "2021-01-01T00:00"
edate = "2021-01-01T02:00"
var = ["Tair"]
shp_crs = 4326
shp_poly_idx = "huc12"
wght_gen_crs = 6931


def test_calculate_weights(get_xarray, get_gdf, get_file_path, get_out_path):
    """Test calculate weights."""
    user_data = UserCatData(
        ds=get_xarray,
        proj_ds=data_crs,
        x_coord=x_coord,
        y_coord=y_coord,
        t_coord=t_coord,
        var=var,
        f_feature=get_gdf,
        proj_feature=shp_crs,
        id_feature=shp_poly_idx,
        period=[sdate, edate],
    )

    tempfile = NamedTemporaryFile()

    wght_gen = WeightGen(
        user_data=user_data,
        method="serial",
        output_file=tempfile,
        weight_gen_crs=wght_gen_crs,
    )

    _wghts = wght_gen.calculate_weights()

    assert isinstance(_wghts, pd.DataFrame)

    tmpdir = TemporaryDirectory()

    agg_gen = AggGen(
        user_data=user_data,
        stat_method="masked_average",
        agg_engine="serial",
        agg_writer="csv",
        weights=tempfile.name,
        out_path=tmpdir.name,
        file_prefix="gm_tmax",
    )

    _ngdf, _vals = agg_gen.calculate_agg()

    assert isinstance(_ngdf, gpd.GeoDataFrame)
    assert isinstance(_vals[0], np.ndarray)

    ofile = get_out_path / tempfile.name
    assert ofile.exists()

    outfile = pd.read_csv(ofile)
    print(outfile.head())



"""Tests for .helper functions."""
import geopandas as gpd
import pandas as pd
import pytest
import xarray as xr

from gdptools.data.odap_cat_data import CatGrids
from gdptools.data.odap_cat_data import CatParams
from gdptools.helpers import get_data_subset_catalog


@pytest.fixture()
def get_gdf() -> gpd.GeoDataFrame:
    """Create GeoDataFrame."""
    return gpd.read_file("./tests/data/hru_1210_epsg5070.shp")


@pytest.fixture()
def get_xarray() -> xr.Dataset:
    """Create xarray Dataset."""
    # return xr.open_dataset("./tests/data/cape_cod_tmax.nc")
    return xr.open_dataset(
        "http://thredds.northwestknowledge.net:8080/thredds/dodsC/agg_met_tmmx_1979_CurrentYear_CONUS.nc"
    )


@pytest.fixture()
def get_file_path(tmp_path):
    """Get temp file path."""
    return tmp_path / "test.csv"


@pytest.fixture()
def get_out_path(tmp_path):
    """Get temp file output path."""
    return tmp_path


data_crs = 4326
x_coord = "lon"
y_coord = "lat"
t_coord = "day"
sdate = "1979-01-01"
edate = "1979-01-07"
var = "daily_maximum_temperature"
shp_crs = 5070
shp_poly_idx = "hru_id_nat"
wght_gen_crs = 6931


@pytest.fixture(scope="module")
def catparam() -> CatParams:
    """Return parameter json."""
    cat_params = "https://mikejohnson51.github.io/opendap.catalog/cat_params.json"
    params = pd.read_json(cat_params)
    _id = "terraclim"  # noqa
    _varname = "aet"  # noqa
    tc = params.query("id == @_id & varname == @_varname")
    return CatParams(**tc.to_dict("records")[0])


@pytest.fixture(scope="module")
def catgrid(catparam) -> CatGrids:
    """Return grid json."""
    cat_grid = "https://mikejohnson51.github.io/opendap.catalog/cat_grids.json"
    grids = pd.read_json(cat_grid)
    _gridid = catparam.grid_id  # noqa
    tc = grids.query("grid_id == @_gridid")
    return CatGrids(**tc.to_dict("records")[0])


@pytest.fixture(scope="module")
def get_gdf_world() -> gpd.GeoDataFrame:
    """Get gdf file with country testing."""
    return gpd.read_file(
        "./tests/data/TM_WORLD_BORDERS_SIMPL-0.3/TM_WORLD_BORDERS_SIMPL-0.3.shp"
    )


@pytest.fixture(scope="module")
def get_begin_date() -> str:
    """Get begin date."""
    return "2005-01-01"


@pytest.fixture(scope="module")
def get_end_date() -> str:
    """Get end date."""
    return "2005-02-01"


@pytest.mark.parametrize(
    "cp,cg,gdf,sd,ed,name",
    [
        (
            "catparam",
            "catgrid",
            "get_gdf_world",
            "get_begin_date",
            "get_end_date",
            "Chile",
        ),
        (
            "catparam",
            "catgrid",
            "get_gdf_world",
            "get_begin_date",
            "get_end_date",
            "Netherlands",
        ),
        (
            "catparam",
            "catgrid",
            "get_gdf_world",
            "get_begin_date",
            "get_end_date",
            "Kenya",
        ),
        (
            "catparam",
            "catgrid",
            "get_gdf_world",
            "get_begin_date",
            "get_end_date",
            "Samoa",
        ),
        (
            "catparam",
            "catgrid",
            "get_gdf_world",
            "get_begin_date",
            "get_end_date",
            "Fiji",
        ),
    ],
)
def test_get_data_subset_catalog(cp, cg, gdf, sd, ed, name, request):
    """Test subset catalog."""
    ds = get_data_subset_catalog(
        cat_params=request.getfixturevalue(cp),
        cat_grid=request.getfixturevalue(cg),
        shp_file=request.getfixturevalue(gdf).query("NAME == @name"),
        begin_date=request.getfixturevalue(sd),
        end_date=request.getfixturevalue(ed),
    )

    assert isinstance(ds, xr.DataArray)



"""OpenDAP Catalog Data classes."""
from __future__ import annotations

from typing import Optional

import numpy as np
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator


class CatParams(BaseModel):
    """Class representing elements of Mike Johnsons OpenDAP catalog params.

    https://mikejohnson51.github.io/opendap.catalog/cat_params.json
    """

    id: Optional[str] = None
    URL: str
    grid_id: Optional[int] = -1
    variable: Optional[str] = None
    varname: str
    long_name: Optional[str] = ...
    T_name: Optional[str] = ...
    duration: Optional[str] = None
    units: Optional[str] = ...
    interval: Optional[str] = None
    nT: Optional[int] = 0  # noqa
    tiled: Optional[str] = None
    model: Optional[str] = None
    ensemble: Optional[str] = None
    scenario: Optional[str] = None

    @validator("grid_id", pre=True, always=True)
    def set_grid_id(cls, v):  # noqa:
        """Convert to int."""
        return int(v)

    @validator("nT", pre=True, always=False)
    def set_nt(cls, v):  # noqa:
        """Convert to int."""
        return 0 if np.isnan(v) else int(v)


class CatGrids(BaseModel):
    """Class representing elements of Mike Johnsons OpenDAP catalog grids.

    https://mikejohnson51.github.io/opendap.catalog/cat_grids.json
    """

    grid_id: Optional[int] = None
    X_name: str
    Y_name: str
    X1: Optional[float] = None
    Xn: Optional[float] = None
    Y1: Optional[float] = None
    Yn: Optional[float] = None
    resX: Optional[float] = None  # noqa
    resY: Optional[float] = None  # noqa
    ncols: Optional[int] = None
    nrows: Optional[int] = None
    proj: str
    toptobottom: int
    tile: Optional[str] = None
    grid_id_1: Optional[str] = Field(None, alias="grid.id")

    @validator("toptobottom", pre=True, always=True)
    def get_toptobottom(cls, v):  # noqa:
        """Convert str to int."""
        return int(v)



"""Statistical Funtions for aggregation."""
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any

import numpy as np
import numpy.typing as npt


@dataclass
class StatsMethod(ABC):
    """Abstract method for gdptools weighted stats."""

    @abstractmethod
    def get_stat(
        self,
    ) -> Any:
        """Abstract weighted method."""


@dataclass
class MAWeightedAverage(StatsMethod):
    """Weighted Average."""

    array: npt.NDArray
    weights: npt.NDArray[np.double]
    def_val: Any

    def get_stat(self) -> Any:
        """Get weighted masked average."""
        try:
            tmp = np.ma.average(self.array, weights=self.weights, axis=1)
        except KeyError:
            numpts = self.weights.shape[0]
            tmp = np.full((numpts), self.def_val)
        return tmp


@dataclass
class WeightedAverage(StatsMethod):
    """Weighted Average."""

    array: npt.NDArray
    weights: npt.NDArray[np.double]
    def_val: Any

    def get_stat(self) -> Any:
        """Get weighted masked average."""
        try:
            tmp = np.average(self.array, weights=self.weights, axis=1)
        except KeyError:
            numpts = self.weights.shape[0]
            tmp = np.full((numpts), self.def_val)
        return tmp



"""Calculate aggregation methods."""
from typing import List
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import Union

import geopandas as gpd
import numpy.typing as npt
import pandas as pd

from gdptools.agg.agg_data_writers import CSVWriter
from gdptools.agg.agg_data_writers import NetCDFWriter
from gdptools.agg.agg_engines import SerialAgg
from gdptools.agg.stats_methods import MAWeightedAverage
from gdptools.agg.stats_methods import WeightedAverage
from gdptools.data.agg_gen_data import AggData
from gdptools.data.user_data import UserData

STATSMETHODS = Literal["masked_average", "average"]
""" List of available stats methods to apply to AggGen class.

masked_average: Accounts for missing gridded values and returns the partial weighted
    value.
average: Returns the weighted avarage.  If a polygon contains a cell value that is
    missing/nan then it will return a nan for that polygon.


Raises:
    TypeError: If supplied value is not one of STATSMETHODS

Returns:
    _type_: str
"""

AGGENGINES = Literal["serial"]
""" List of aggregation methods.

serial: performes weighted-area aggregation by iterating through polygons.

TODO: parallel

Raises:
    TypeError: If supplied attribute is not one of AGGENGINES.

Returns:
    _type_: str
"""

AGGWRITERS = Literal["csv", "netcdf"]
""" List of available writers applied to the aggregation.

csv: Output data in csv format.
netcdf: Output data in netcdf format.

Raises:
    TypeError: If supplied attribute is not one of AGGWRITERS.

Returns:
    _type_: str
"""


class AggGen:
    """Class for aggregating grid-to-polygons."""

    def __init__(
        self,
        user_data: UserData,
        stat_method: STATSMETHODS,
        agg_engine: AGGENGINES,
        agg_writer: AGGWRITERS,
        weights: Union[str, pd.DataFrame],
        out_path: str,
        file_prefix: str,
        append_date: Optional[bool] = False,
    ) -> None:
        """Initialize AggGen.

        Args:
            user_data (UserData): _description_
            stat_method (STATSMETHODS): _description_
            agg_engine (AGGENGINES): _description_
            agg_writer (AGGWRITERS): _description_
            weights (Union[str, pd.DataFrame]): _description_
            out_path (str): _description_
            file_prefix (str): _description_
            append_date (Optional[bool], optional): _description_. Defaults to False.

        Raises:
            TypeError: _description_
            TypeError: _description_
            TypeError: _description_
        """
        self._user_data = user_data
        self._stat_method = stat_method
        self._agg_engine = agg_engine
        self._agg_writer = agg_writer
        self._weights = weights
        self._out_path = out_path
        self._file_prefix = file_prefix
        self._append_date = append_date
        self._agg_data = None

        if self._stat_method == "masked_average":
            self._stat = MAWeightedAverage
        elif self._stat_method == "average":
            self._stat = WeightedAverage
        else:
            raise TypeError(f"stat_method: {self._stat_method} not in {STATSMETHODS}")

        if self._agg_engine == "serial":
            self.agg = SerialAgg
        else:
            raise TypeError(f"agg_engine: {self._agg_engine} not in {AGGENGINES}")

        if self._agg_writer == "csv":
            self.__writer = CSVWriter
        elif self._agg_writer == "netcdf":
            self.__writer = NetCDFWriter
        else:
            raise TypeError(f"agg_writer: {self._agg_writer} not in {AGGWRITERS}")

    def calculate_agg(self) -> Tuple[gpd.GeoDataFrame, List[npt.NDArray]]:
        """Calculate aggregations.

        Returns:
            Tuple[gpd.GeoDataFrame, List[npt.NDArray]]: _description_
        """
        self._agg_data, new_gdf, agg_vals = self.agg().calc_agg_from_dictmeta(
            user_data=self._user_data,
            weights=self._weights,
            stat=self._stat,
        )
        self.__writer().save_file(
            agg_data=self._agg_data,
            feature=new_gdf,
            vals=agg_vals,
            p_out=self._out_path,
            file_prefix=self._file_prefix,
            append_date=self._append_date,
        )

        return new_gdf, agg_vals

    @property
    def agg_data(self) -> AggData:
        """Return agg_data."""
        return self._agg_data



"""Top-level package for pygeoapi plugin: Gdptools."""
import logging

from .agg_gen import AggGen
from .data.user_data import ODAPCatData
from .data.user_data import UserCatData
from .data.user_data import UserTiffData
from .weight_gen import WeightGen
from .zonal_gen import ZonalGen

__author__ = "Richard McDonald"
__email__ = "rmcd@usgs.gov"
__version__ = "0.0.29"

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "WeightGen",
    "AggGen",
    "ZonalGen",
    "ODAPCatData",
    "UserCatData",
    "UserTiffData",
]



"""Method for saving aggregation data."""
import warnings
from abc import ABC
from abc import abstractmethod
from datetime import datetime
from pathlib import Path
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import geopandas as gpd
import numpy as np
import numpy.typing as npt
import pandas as pd
import pyproj
import xarray as xr
from shapely.geometry import Point

from gdptools.data.agg_gen_data import AggData


class AggDataWriter(ABC):
    """Abstract writer Template method."""

    def save_file(
        self: "AggDataWriter",
        agg_data: List[AggData],
        feature: gpd.GeoDataFrame,
        vals: List[npt.NDArray[Union[np.integer, np.double]]],
        p_out: str,
        file_prefix: str,
        append_date: Optional[bool] = False,
    ) -> None:
        """Abstract Template method for writing aggregation data.

        Args:
            agg_data (List[AggData]): _description_
            feature (gpd.GeoDataFrame): _description_
            vals (List[npt.NDArray[Union[int, np.double]]]): _description_
            p_out (str): _description_
            file_prefix (str): _description_
            append_date (Optional[bool], optional): _description_. Defaults to False.

        Raises:
            FileNotFoundError: _description_
        """
        self.agg_data = agg_data
        self.feature = feature
        self.vals = vals
        self.append_date = append_date
        self.fdate = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
        self.outpath = Path(p_out)
        if not self.outpath.exists():
            raise FileNotFoundError(f"Path: {p_out} does not exist")
        if self.append_date:
            self.fname = f"{self.fdate}_{file_prefix}"
        else:
            self.fname = f"{file_prefix}"

        self.create_out_file()

    @abstractmethod
    def create_out_file(self) -> None:
        """Abstract method for writing aggregation data."""
        pass


class CSVWriter(AggDataWriter):
    """Class for writing csv files."""

    def create_out_file(self) -> None:
        """Method for writing csv files."""
        # for idx in range(len(self.agg_data)):
        for idx, (_key, value) in enumerate(self.agg_data.items()):
            gdf = self.feature
            # gdf_idx = self.agg_data[idx].id_feature
            param_values = value.cat_param
            t_coord = param_values.T_name
            units = param_values.units
            varname = param_values.varname
            time = value.da.coords[t_coord].values
            # units = self.agg_data[idx].param_dict

            df_key = pd.DataFrame(data=self.vals[idx], columns=gdf.index.T.values)

            df_key.insert(0, "units", [units] * df_key.shape[0])
            df_key.insert(0, "varname", [varname] * df_key.shape[0])
            df_key.insert(0, "time", time)

            if idx == 0:
                df = df_key
            else:
                pd.concat([df, df_key])
        df.reset_index(inplace=True)
        path_to_file = self.outpath / f"{self.fname}.csv"
        print(f"Saving csv file to {path_to_file}")
        df.to_csv(path_to_file)


class NetCDFWriter(AggDataWriter):
    """NetCDF writer class."""

    def create_out_file(self) -> None:
        """Create NetCDF file from aggregation data."""
        # Suppres UserWarning from centroid calc - Here lat/lon centroid are
        # a convenience method.
        warnings.filterwarnings(action="ignore", category=UserWarning)
        dataset = []
        # for idx in range(len(self.agg_data)):
        for idx, (_key, value) in enumerate(self.agg_data.items()):
            gdf = self.feature
            gdf_idx = value.id_feature
            # param_values = list(self.agg_data[idx].param_dict.values())[idx]
            param_values = value.cat_param
            t_coord = param_values.T_name
            v_units = param_values.units
            v_varname = param_values.varname
            v_long_name = param_values.long_name
            time = value.da.coords[t_coord].values
            locs = gdf.index.values

            def getxy(pt: Point) -> Tuple[np.double, np.double]:
                """Return x y components of point."""
                return pt.x, pt.y

            centroid_series = gdf.geometry.centroid
            tlon, tlat = [
                list(t) for t in zip(*map(getxy, centroid_series))  # noqa B905
            ]
            crs_meta = pyproj.CRS(gdf.crs).to_cf()

            dsn = xr.Dataset(
                data_vars={
                    v_varname: (
                        ["time", gdf_idx],
                        self.vals[idx],
                        dict(
                            units=v_units,
                            long_name=v_long_name,
                            coordinates="time lat lon",
                            grid_mapping="crs",
                        ),
                    ),
                    "crs": (["one"], np.ones((1), dtype=np.double), crs_meta),
                },
                coords={
                    "time": time,
                    gdf_idx: ([gdf_idx], locs, {"feature_id": gdf_idx}),
                    "lat": (
                        [gdf_idx],
                        tlat,
                        {
                            "long_name": "Latitude of HRU centroid",
                            "units": "degrees_north",
                            "standard_name": "latitude",
                            "axis": "Y",
                        },
                    ),
                    "lon": (
                        [gdf_idx],
                        tlon,
                        {
                            "long_name": "Longitude of HRU centroid",
                            "units": "degrees_east",
                            "standard_name": "longitude",
                            "axis": "X",
                        },
                    ),
                },
            )

            dataset.append(dsn)
        # crs_cf = pyproj.CRS(gdf.crs).to_cf()
        # da = xr.DataArray(
        #     data=np.ones((1), dtype=np.double),
        #     attrs=crs_cf
        # )
        # data_array.append(da)
        ds = xr.merge(dataset)
        # ds = ds.assign(
        #     crs=(["one"], np.ones((1), dtype=np.double))
        # )
        # ds = ds.assign_coords(lat=tlat)
        # ds.lat.attrs = {
        #     "long_name": "Latitude of HRU centroid",
        #     "units": "degrees_north",
        #     "standard_name": "latitude",
        #     "axis": "Y"}

        # ds = ds.assign_coords(lon=tlon)
        # ds.lon.attrs = {
        #     "long_name": "Longitude of HRU centroid",
        #     "units": "degrees_east",
        #     "standard_name": "longitude",
        #     "axis": "X"}
        # ds.crs.attrs = crs_cf
        ds.attrs = {
            "Conventions": "CF-1.8",
            "featureType": "timeSeries",
            "history": (
                f"{self.fdate} Original filec created  by gdptools package: "
                "https://code.usgs.gov/wma/nhgf/toolsteam/gdptools \n"
            ),
        }
        path_to_file = self.outpath / f"{self.fname}.nc"
        print(f"Saving netcdf file to {path_to_file}")
        ds.to_netcdf(path_to_file)



"""Engines for calculated zonal stats based on non-area-weighted statistics."""
import os
import time
from abc import ABC
from abc import abstractmethod
from typing import Optional

import geopandas as gpd
import numpy as np
import pandas as pd
from joblib import delayed
from joblib import Parallel
from joblib import parallel_backend

from gdptools.data.agg_gen_data import AggData
from gdptools.data.user_data import UserData
from gdptools.helpers import build_subset_tiff_da
from gdptools.utils import _get_shp_bounds_w_buffer
from gdptools.weights.calc_weight_engines import _make_valid


class ZonalEngine(ABC):
    """Base class for zonal stats engines."""

    def calc_zonal_from_aggdata(
        self, user_data: UserData, categorical: Optional[bool] = False
    ) -> pd.DataFrame:
        """calc_zonal_from_aggdata Template method for calculated zonal stats.

        _extended_summary_

        Args:
            user_data (UserData): _description_
            categorical (Optional[bool], optional): _description_. Defaults to False.

        Returns:
            pd.DataFrame: _description_
        """
        self._user_data = user_data
        self._categorical = categorical

        return self.zonal_stats()

    @abstractmethod
    def zonal_stats(self) -> pd.DataFrame:
        """Abstract method for calculating zonal stats."""
        pass


class ZonalEngineSerial(ZonalEngine):
    """Serial zonal stats engine."""

    def zonal_stats(self) -> pd.DataFrame:
        """zonal_stats Calculate zonal stats serially.

        _extended_summary_

        Returns:
            pd.DataFrame: _description_
        """
        zvars = self._user_data.get_vars()
        tstrt = time.perf_counter()
        agg_data: AggData = self._user_data.prep_agg_data(zvars[0])
        tend = time.perf_counter()
        print(f"data prepped for zonal in {tend - tstrt:0.4f} seconds")
        ds_ss = agg_data.da

        if self._categorical:
            d_categories = list(pd.Categorical(ds_ss.values.flatten()).categories)

        tstrt = time.perf_counter()
        lon, lat = np.meshgrid(
            ds_ss[agg_data.cat_grid.X_name].values,
            ds_ss[agg_data.cat_grid.Y_name].values,
        )
        lat_flat = lat.flatten()
        lon_flat = lon.flatten()
        ds_vals = ds_ss.values.flatten()
        df_points = pd.DataFrame(
            {
                "index": np.arange(len(lat_flat)),
                "vals": ds_vals,
                "lat": lat_flat,
                "lon": lon_flat,
            }
        )
        try:
            fill_val = ds_ss._FillValue
            df_points_filt = df_points[df_points.vals != fill_val]
        except Exception:
            df_points_filt = df_points
        # df_points_filt = df_points[df_points.vals >= ds_min]
        source_df = gpd.GeoDataFrame(
            df_points_filt,
            geometry=gpd.points_from_xy(df_points_filt.lon, df_points_filt.lat),
        )
        tend = time.perf_counter()
        print(f"converted tiff to points in {tend - tstrt:0.4f} seconds")
        source_df.set_crs(agg_data.cat_grid.proj, inplace=True)
        target_df = agg_data.feature.to_crs(agg_data.cat_grid.proj)
        target_df = _make_valid(target_df)
        target_df.reset_index()
        target_df_keys = target_df[agg_data.id_feature].values
        tstrt = time.perf_counter()
        ids_tgt, ids_src = source_df.sindex.query_bulk(
            target_df.geometry, predicate="contains"
        )
        tend = time.perf_counter()
        print(f"overlaps calculated in {tend - tstrt:0.4f} seconds")
        if self._categorical:

            val_series = pd.Categorical(
                source_df["vals"].iloc[ids_src], categories=d_categories
            )

            agg_df = pd.DataFrame(
                {
                    agg_data.id_feature: target_df[agg_data.id_feature]
                    .iloc[ids_tgt]
                    .values
                }
            )
            agg_df["vals"] = val_series
            tstrt = time.perf_counter()
            stats = agg_df.groupby(agg_data.id_feature)["vals"].describe(
                include=["category"]
            )
            tend = time.perf_counter()
            print(f"categorical zonal stats calculated in {tend - tstrt:0.4f} seconds")
        else:
            agg_df = pd.DataFrame(
                {
                    agg_data.id_feature: target_df[agg_data.id_feature]
                    .iloc[ids_tgt]
                    .values,
                    "vals": source_df["vals"].iloc[ids_src],
                }
            )
            tstrt = time.perf_counter()
            stats = agg_df.groupby(agg_data.id_feature)["vals"].describe()
            # stats.insert(-1, "sum", agg_df.groupby(agg_data.id_feature).sum())
            stats["sum"] = agg_df.groupby(agg_data.id_feature).sum()
            # stats.set_index(agg_data.id_feature)
            tend = time.perf_counter()
            print(f"zonal stats calculated in {tend - tstrt:0.4f} seconds")

        tstrt = time.perf_counter()
        stats_inds = stats.index

        missing = np.setdiff1d(target_df_keys, stats_inds)
        target_df_stats = target_df.loc[
            target_df[agg_data.id_feature].isin(list(stats_inds))
        ]
        target_df_missing = target_df.loc[
            target_df[agg_data.id_feature].isin(list(missing))
        ]
        nearest = target_df_stats.sindex.nearest(
            target_df_missing.geometry, return_all=False
        )
        print(nearest)
        print(f"number of missing values: {len(missing)}")
        stats_missing = stats.iloc[nearest[1]]
        stats_missing.index = missing
        stats_tot = pd.concat([stats, stats_missing])
        stats_tot.index.name = agg_data.id_feature
        tend = time.perf_counter()
        print(
            f"fill missing values with nearest neighbors in {tend - tstrt:0.4f} seconds"
        )

        return stats_tot


class ZonalEngineParallel(ZonalEngine):
    """Parallel zonal stats engine."""

    def zonal_stats(self) -> pd.DataFrame:
        """zonal_stats Calculate zonal stats serially.

        _extended_summary_

        Returns:
            pd.DataFrame: _description_
        """
        n_jobs = os.cpu_count() - 1
        # n_jobs = 2
        zvars = self._user_data.get_vars()
        tstrt = time.perf_counter()
        agg_data: AggData = self._user_data.prep_agg_data(zvars[0])
        tend = time.perf_counter()
        print(f"data prepped for zonal in {tend - tstrt:0.4f} seconds")
        ds_ss = agg_data.da

        if self._categorical:
            d_categories = list(pd.Categorical(ds_ss.values.flatten()).categories)
        else:
            d_categories = []

        target_df = agg_data.feature.set_crs(agg_data.cat_grid.proj, inplace=True)
        target_df = _make_valid(target_df)
        # target_df.reset_index(inplace=True)
        target_df_keys = target_df[agg_data.id_feature].values
        to_workers = _chunk_dfs(
            target_df,
            ds_ss,
            agg_data.cat_grid.X_name,
            agg_data.cat_grid.Y_name,
            agg_data.cat_grid.proj,
            agg_data.cat_grid.toptobottom,
            self._categorical,
            agg_data.id_feature,
            d_categories,
            n_jobs=n_jobs,
        )

        tstrt = time.perf_counter()
        worker_out = self.get_stats_parallel(n_jobs, to_workers)
        for i in range(len(worker_out)):
            if i == 0:
                stats = worker_out[i]
                print(type(stats))
            else:
                stats = pd.concat([stats, worker_out[i]])
        stats.set_index(agg_data.id_feature, inplace=True)
        # stats = pd.concat(worker_out)
        tend = time.perf_counter()
        print(f"Parallel calculation of zonal stats in {tend - tstrt:0.04} seconds")

        tstrt = time.perf_counter()
        stats_inds = stats.index

        missing = np.setdiff1d(target_df_keys, stats_inds)
        if len(missing) <= 0:
            return stats
        target_df_stats = target_df.loc[
            target_df[agg_data.id_feature].isin(list(stats_inds))
        ]
        target_df_missing = target_df.loc[
            target_df[agg_data.id_feature].isin(list(missing))
        ]
        nearest = target_df_stats.sindex.nearest(
            target_df_missing.geometry, return_all=False
        )
        print(nearest)
        print(f"number of missing values: {len(missing)}")
        stats_missing = stats.iloc[nearest[1]]
        stats_missing.index = missing
        stats_tot = pd.concat([stats, stats_missing])
        stats_tot.index.name = agg_data.id_feature
        tend = time.perf_counter()
        print(
            f"fill missing values with nearest neighbors in {tend - tstrt:0.4f} seconds"
        )

        return stats_tot

    def get_stats_parallel(self, n_jobs, to_workers):
        """get_stats_parallel -Parallel engine.

        _extended_summary_

        Args:
            n_jobs (int): Number of processors
            to_workers (list): List of chunked dataframe and addtional parameters

        Returns:
            _type_: _description_
        """
        with parallel_backend("loky", inner_max_num_threads=1):
            worker_out = Parallel(n_jobs=n_jobs)(
                delayed(_get_stats_on_chunk)(*chunk_pair) for chunk_pair in to_workers
            )
        return worker_out


def _get_stats_on_chunk(
    geom, df, x_name, y_name, proj, toptobottom, categorical, id_feature, cats
):
    num_features = len(geom.index)
    comp_stats = []
    for index in range(num_features):
        target_df = geom.iloc[[index]]
        feat_bounds = _get_shp_bounds_w_buffer(
            gdf=target_df, ds=df, crs=proj, lon=x_name, lat=y_name
        )

        subset_dict = build_subset_tiff_da(
            bounds=feat_bounds, xname=x_name, yname=y_name, toptobottom=toptobottom
        )
        ds_ss = df.sel(**subset_dict)
        lon, lat = np.meshgrid(
            ds_ss[x_name].values,
            ds_ss[y_name].values,
        )
        lat_flat = lat.flatten()
        lon_flat = lon.flatten()
        ds_vals = ds_ss.values.flatten()
        df_points = pd.DataFrame(
            {
                "index": np.arange(len(lat_flat)),
                "vals": ds_vals,
                "lat": lat_flat,
                "lon": lon_flat,
            }
        )
        try:
            fill_val = ds_ss._FillValue
            df_points = df_points[df_points.vals != fill_val]
        except Exception:
            df_points = df_points

        source_df = gpd.GeoDataFrame(
            df_points,
            geometry=gpd.points_from_xy(df_points.lon, df_points.lat),
            crs=proj,
        )

        ids_src = source_df.sindex.query(
            target_df.geometry.values[0], predicate="contains"
        )
        if categorical:

            val_series = pd.Categorical(
                source_df["vals"].iloc[ids_src], categories=cats
            )

            agg_df = pd.DataFrame(data={"vals": val_series})
            stats = agg_df.describe(include=["category"]).T
        else:
            agg_df = pd.DataFrame(
                data={
                    "vals": source_df["vals"].iloc[ids_src].values,
                },
                # index=[target_df[id_feature].values[0]]
            )
            stats = agg_df.describe().T
            stats["sum"] = agg_df["vals"].sum()

        stats[id_feature] = target_df[id_feature].values[0]
        if stats["count"].values[0] <= 0:
            continue
        comp_stats.append(stats)

    return pd.concat(comp_stats)


def _chunk_dfs(
    geoms_to_chunk, df, x_name, y_name, proj, ttb, categorical, id_feature, cats, n_jobs
):
    """Chunk dataframes for parallel processing."""
    start = 0
    chunk_size = geoms_to_chunk.shape[0] // n_jobs + 1

    # if length < chunk_size:
    #     yield \
    #         geoms_to_chunk[:], \
    #         df, \
    #         x_name, \
    #         y_name, \
    #         proj, \
    #         ttb, \
    #         categorical, \
    #         id_feature, \
    #         cats
    #     return

    # while start + chunk_size <= length:
    #     yield \
    #         geoms_to_chunk[start:(chunk_size + start)], \
    #         df, \
    #         x_name, \
    #         y_name, \
    #         proj, \
    #         ttb, \
    #         categorical, \
    #         id_feature, \
    #         cats
    #     print(start, (chunk_size + start))
    #     start += chunk_size

    # if start < length:
    #     yield \
    #         geoms_to_chunk[start:], \
    #         df, \
    #         x_name, \
    #         y_name, \
    #         proj, \
    #         ttb, \
    #         categorical, \
    #         id_feature, \
    #         cats
    #     print(start)
    #     start += chunk_size
    #     return

    # chunk_size = geoms_to_chunk.shape[0] // n_jobs + 1
    for i in range(n_jobs):
        start = i * chunk_size
        yield geoms_to_chunk.iloc[
            start : start + chunk_size
        ], df, x_name, y_name, proj, ttb, categorical, id_feature, cats



"""Typeguard support for tests."""
import pytest

try:
    import typeguard  # noqa: F401
except ImportError:
    has_typeguard = False
else:
    has_typeguard = True


skip_if_typeguard = pytest.mark.skipif(
    has_typeguard,
    reason="Broken if Typeguard is enabled",
)



"""Dataclasses used in aggregation."""
from dataclasses import dataclass
from typing import List

from geopandas import GeoDataFrame
from xarray import DataArray

from gdptools.data.odap_cat_data import CatGrids
from gdptools.data.odap_cat_data import CatParams


@dataclass(repr=True)
class AggData(object):
    """AggData is a convenience container of data necessary for aggregation.

    Data provided in one of UserData inherited classes will be prepped for aggregation,
    including subsetting the gridded data by the features bounding box, and by the time-
    period selected.  In addition if the gridded data is defined between 0-360 degrees
    longitude it will be shifted to -180 - 180 degrees.  For each variable in the
    user_data attribute of either the WeightGen or AggGen classes, a dict of
    {var: AggData} will be generated in the AggGen.calculate_agg() method.

    Args:
        variable (str): Variable name.
        cat_param (CatParams): Catparams data class containing parameter metadata.
        cat_grid (CatGrids): CatGrids data class containing grid metadata.
        da: (DataArray): The spatially and temporally subsetted variable DataArray.
        feature (GeoDataFrame): The user-supplied feature file represented as a
            GeoDataFrame.
        id_feature (str): The feature id (column header) in the GeoDataFrame.
        period (List[str]): A list of dates representing the starting and ending date to
            process.
    """

    variable: str
    cat_param: CatParams
    cat_grid: CatGrids
    da: DataArray
    feature: GeoDataFrame
    id_feature: str
    period: List[str]



"""Data classes."""
from __future__ import annotations

from typing import Any

import xarray as xr
from attrs import define
from attrs import field
from pyproj.crs import CRS


def check_xname(instance, attribute, value):
    """Validate xname."""
    if value not in instance.ds.coords:
        raise ValueError(f"xname:{value} not in {instance.ds.coords}")


def check_yname(instance, attribute, value):
    """Validate yname."""
    if value not in instance.ds.coords:
        raise ValueError(f"yname:{value} not in {instance.ds.coords}")


def check_band(instance, attribute, value):
    """Validate band name."""
    if value not in instance.ds.coords:
        raise ValueError(f"band:{value} not in {instance.ds.coords}")


def check_crs(instance, attribute, value):
    """Validate crs."""
    crs = CRS.from_user_input(value)
    if not isinstance(crs, CRS):
        raise ValueError(f"crs:{crs} not in valid")


@define(kw_only=True)
class TiffAttributes(object):
    """Tiff qttributes data class."""

    varname: str
    xname: str = field(validator=check_xname)
    yname: str = field(validator=check_yname)
    bname: str = field(validator=check_band)
    band: int = field()
    toptobottom: bool = field(init=False)
    crs: Any = field(validator=check_crs)
    categorical: bool = field()
    ds: xr.DataArray

    def __attrs_post_init__(self):
        """Generate toptobottom."""
        yy = self.ds.coords[self.yname].values
        self.toptobottom = yy[0] <= yy[-1]



"""Testing ancillary functions."""
from typing import Any

import geopandas as gpd
import numpy as np
import numpy.typing as npt
import pandas as pd
import pytest
import xarray as xr
from pyproj.crs import CRS
from pyproj.exceptions import CRSError

from gdptools.data.odap_cat_data import CatGrids
from gdptools.data.odap_cat_data import CatParams
from gdptools.utils import _check_for_intersection
from gdptools.utils import _get_cells_poly
from gdptools.utils import _get_crs
from gdptools.utils import _get_data_via_catalog
from gdptools.utils import _get_shp_file


@pytest.mark.parametrize(
    "crs",
    [
        "epsg:4326",
        4326,
        "+proj=longlat +a=6378137 +f=0.00335281066474748 +pm=0 +no_defs",
    ],
)
def test__get_crs(crs: Any) -> None:
    """Test the get_crs function."""
    crs = _get_crs(crs)
    assert isinstance(crs, CRS)


@pytest.mark.parametrize(
    "crs",
    [
        "espg:4326",
        43,
        "+a=6378137 +f=0.00335281066474748 +pm=0 +no_defs",
    ],
)
def test__get_bad_crs(crs: Any) -> None:
    """Test the get_crs function."""
    with pytest.raises(CRSError):
        crs = _get_crs(crs)


@pytest.fixture
def catparam() -> CatParams:
    """Return parameter json."""
    cat_params = "https://mikejohnson51.github.io/opendap.catalog/cat_params.json"
    params = pd.read_json(cat_params)
    _id = "gridmet"  # noqa
    _varname = "daily_maximum_temperature"  # noqa
    tc = params.query("id == @_id & varname == @_varname")
    return CatParams(**tc.to_dict("records")[0])


@pytest.fixture
def catgrid() -> CatGrids:
    """Return grid json."""
    cat_grid = "https://mikejohnson51.github.io/opendap.catalog/cat_grids.json"
    grids = pd.read_json(cat_grid)
    _gridid = 176  # noqa
    tc = grids.query("grid_id == @_gridid")
    return CatGrids(**tc.to_dict("records")[0])


@pytest.fixture
def gdf() -> gpd.GeoDataFrame:
    """Create xarray dataset."""
    return gpd.read_file("./tests/data/hru_1210_epsg5070.shp")


@pytest.fixture
def is_degrees(gdf, catparam, catgrid) -> bool:
    """Check if coords are in degrees."""
    is_intersect, is_degrees, is_0_360 = _check_for_intersection(
        cat_params=catparam, cat_grid=catgrid, gdf=gdf
    )
    return is_degrees


@pytest.fixture
def bounds(gdf, catgrid, is_degrees) -> npt.NDArray[np.double]:
    """Get bounds."""
    gdf, bounds = _get_shp_file(gdf, catgrid, is_degrees)
    return bounds


@pytest.fixture
def xarray(catparam, catgrid, bounds) -> xr.Dataset:
    """Create xarray dataset."""
    return _get_data_via_catalog(catparam, catgrid, bounds, "2020-01-01")


def test__get_cells_poly(catparam, catgrid, bounds) -> None:
    """Test _get_cells_poly."""
    ds = _get_data_via_catalog(catparam, catgrid, bounds, "2020-01-01")
    print(ds)
    assert isinstance(ds, xr.DataArray)
    gdf = _get_cells_poly(xr_a=ds, x="lon", y="lat", crs_in=4326)
    assert isinstance(gdf, gpd.GeoDataFrame)



"""Test WghtGen and AggGen classes."""
from pathlib import Path
from tempfile import NamedTemporaryFile
from tempfile import TemporaryDirectory
from typing import Dict

import geopandas as gpd
import numpy as np
import numpy.typing as npt
import pandas as pd
import pytest
import xarray as xr

from gdptools import AggGen
from gdptools import ODAPCatData
from gdptools import UserCatData
from gdptools import WeightGen

gm_vars = ["aet", "pet", "PDSI"]


@pytest.fixture
def aet_ma_average() -> npt.NDArray[np.double]:
    """Test values for masked-average calculation."""
    return np.asarray(
        [
            [3.1605408],
            [0.0121277],
            [41.983784],
            [68.91504],
            [81.72351],
            [92.60617],
            [48.96107],
            [37.30392],
            [27.883911],
            [64.65818],
            [52.897045],
            [1.6390916],
        ]
    )


@pytest.fixture
def aet_average() -> npt.NDArray[np.double]:
    """Test values for average calculation."""
    return np.asarray(
        [
            [3.1551743],
            [0.0121071],
            [41.9124973],
            [68.7980207],
            [81.5847434],
            [92.4489252],
            [48.8779355],
            [37.2405776],
            [27.8365651],
            [64.5483926],
            [52.8072282],
            [1.63630845],
        ]
    )


@pytest.fixture
def param_dict(vars=gm_vars) -> Dict:
    """Return parameter json."""
    cat_params = "https://mikejohnson51.github.io/opendap.catalog/cat_params.json"
    params = pd.read_json(cat_params)
    _id = "terraclim"  # noqa
    var_params = [
        params.query(
            "id == @_id & variable == @_var", local_dict={"_id": _id, "_var": _var}
        ).to_dict(orient="records")[0]
        for _var in vars
    ]
    return dict(zip(vars, var_params))  # noqa B905


@pytest.fixture
def grid_dict(vars=gm_vars) -> Dict:
    """Return grid json."""
    cat_grid = "https://mikejohnson51.github.io/opendap.catalog/cat_grids.json"
    grids = pd.read_json(cat_grid)
    _gridid = 116  # noqa
    var_grid = [
        grids.query(
            "grid_id == @_gridid", local_dict={"_gridid": _gridid, "_var": _var}
        ).to_dict(orient="records")[0]
        for _var in vars
    ]
    return dict(zip(vars, var_grid))  # noqa B905


@pytest.fixture
def gdf() -> gpd.GeoDataFrame:
    """Create xarray dataset."""
    return gpd.read_file("./tests/data/hru_1210_epsg5070.shp")


@pytest.fixture
def poly_idx() -> str:
    """Return poly_idx."""
    return "hru_id_nat"


@pytest.fixture
def wght_gen_proj() -> int:
    """Return wght gen projection."""
    return 6931


def test_odapcatdata_p(
    param_dict, grid_dict, gdf, poly_idx, wght_gen_proj, aet_ma_average, aet_average
) -> None:
    """Test ODAPCatData."""
    user_data = ODAPCatData(
        param_dict=param_dict,
        grid_dict=grid_dict,
        f_feature=gdf,
        id_feature=poly_idx,
        period=["1980-01-01", "1980-12-31"],
    )
    tmpfile = NamedTemporaryFile()

    gen_weights = WeightGen(
        user_data=user_data,
        method="parallel",
        output_file=tmpfile.name,
        weight_gen_crs=wght_gen_proj,
    )

    _wghts = gen_weights.calculate_weights()  # noqa: F841

    tmpdir = TemporaryDirectory()
    gen_agg = AggGen(
        user_data=user_data,
        stat_method="masked_average",
        agg_engine="serial",
        agg_writer="netcdf",
        weights=tmpfile.name,
        out_path=tmpdir.name,
        file_prefix="test_agg_gen",
    )

    _ngdf, nvals = gen_agg.calculate_agg()
    file = Path(tmpdir.name) / "test_agg_gen.nc"
    assert file.exists()

    np.testing.assert_allclose(aet_ma_average, nvals[0], rtol=1e-4, verbose=True)

    tmpdir2 = TemporaryDirectory()
    gen_agg = AggGen(
        user_data=user_data,
        stat_method="average",
        agg_engine="serial",
        agg_writer="csv",
        weights=tmpfile.name,
        out_path=tmpdir2.name,
        file_prefix="test_agg_gen",
    )

    test_arr = np.asarray(
        [
            [3.1551743],
            [0.0121071],
            [41.9124973],
            [68.7980207],
            [81.5847434],
            [92.4489252],
            [48.8779355],
            [37.2405776],
            [27.8365651],
            [64.5483926],
            [52.8072282],
            [1.63630845],
        ]
    )
    _ngdf, nvals2 = gen_agg.calculate_agg()
    file = Path(tmpdir2.name) / "test_agg_gen.csv"
    assert file.exists()
    print(nvals2[0])
    np.testing.assert_allclose(test_arr, nvals2[0], rtol=1e-4, verbose=True)


def test_odapcatdata(
    param_dict, grid_dict, gdf, poly_idx, wght_gen_proj, tvars=gm_vars
) -> None:
    """Test ODAPCatData."""
    user_data = ODAPCatData(
        param_dict=param_dict,
        grid_dict=grid_dict,
        f_feature=gdf,
        id_feature=poly_idx,
        period=["1980-01-01", "1980-12-31"],
    )
    tmpfile = NamedTemporaryFile()

    gen_weights = WeightGen(
        user_data=user_data,
        method="serial",
        output_file=tmpfile.name,
        weight_gen_crs=wght_gen_proj,
    )

    _wghts = gen_weights.calculate_weights()  # noqa: F841

    tmpdir = TemporaryDirectory()

    gen_agg = AggGen(
        user_data=user_data,
        stat_method="masked_average",
        agg_engine="serial",
        agg_writer="netcdf",
        weights=tmpfile.name,
        out_path=tmpdir.name,
        file_prefix="test_agg_gen",
    )

    _ngdf, nvals = gen_agg.calculate_agg()

    tmpdir2 = TemporaryDirectory()
    gen_agg2 = AggGen(
        user_data=user_data,
        stat_method="masked_average",
        agg_engine="serial",
        agg_writer="csv",
        weights=tmpfile.name,
        out_path=tmpdir2.name,
        file_prefix="test_agg_gen",
    )

    _ngdf2, _nvals2 = gen_agg2.calculate_agg()

    test_arr = np.asarray(
        [
            [3.1605408],
            [0.0121277],
            [41.983784],
            [68.91504],
            [81.72351],
            [92.60617],
            [48.96107],
            [37.30392],
            [27.883911],
            [64.65818],
            [52.897045],
            [1.6390916],
        ]
    )

    np.testing.assert_allclose(test_arr, nvals[0], rtol=1e-4, verbose=True)

    file = Path(tmpdir.name) / "test_agg_gen.nc"

    assert file.exists()


@pytest.fixture()
def get_gdf() -> gpd.GeoDataFrame:
    """Create GeoDataFrame."""
    return gpd.read_file("./tests/data/hru_1210_epsg5070.shp")


@pytest.fixture()
def get_xarray() -> xr.Dataset:
    """Create xarray Dataset."""
    # return xr.open_dataset("./tests/data/cape_cod_tmax.nc")
    return xr.open_dataset(
        "http://thredds.northwestknowledge.net:8080/thredds/dodsC/agg_met_tmmx_1979_CurrentYear_CONUS.nc"
    )


@pytest.fixture()
def get_file_path(tmp_path):
    """Get temp file path."""
    return tmp_path / "test.csv"


@pytest.fixture()
def get_out_path(tmp_path):
    """Get temp file output path."""
    return tmp_path


data_crs = 4326
x_coord = "lon"
y_coord = "lat"
t_coord = "day"
sdate = "1979-01-01"
edate = "1979-01-07"
var = "daily_maximum_temperature"
shp_crs = 5070
shp_poly_idx = "hru_id_nat"
wght_gen_crs = 6931


def test_usercatdata(get_xarray, get_gdf, get_file_path, get_out_path) -> None:
    """Test UserCatData."""
    user_data = UserCatData(
        ds=get_xarray,
        proj_ds=data_crs,
        x_coord=x_coord,
        y_coord=y_coord,
        t_coord=t_coord,
        var=[var],
        f_feature=get_gdf,
        proj_feature=shp_crs,
        id_feature=shp_poly_idx,
        period=[sdate, edate],
    )

    tmpfile = NamedTemporaryFile()
    weight_gen = WeightGen(
        user_data=user_data,
        method="serial",
        output_file=tmpfile.name,
        weight_gen_crs=wght_gen_crs,
    )

    wghts = weight_gen.calculate_weights()

    weight_gen2 = WeightGen(
        user_data=user_data,
        method="parallel",
        output_file=tmpfile.name,
        weight_gen_crs=wght_gen_crs,
    )

    _wghts2 = weight_gen2.calculate_weights()  # noqa: F841

    tmpdir = TemporaryDirectory()

    agg_gen = AggGen(
        user_data=user_data,
        stat_method="average",
        agg_engine="serial",
        agg_writer="csv",
        weights=wghts,
        out_path=tmpdir.name,
        file_prefix="test_agg_gen_2",
    )

    _ngdf, _nvals = agg_gen.calculate_agg()
    print(_nvals[0])

    tmpdir2 = TemporaryDirectory()
    agg_gen = AggGen(
        user_data=user_data,
        stat_method="masked_average",
        agg_engine="serial",
        agg_writer="netcdf",
        weights=wghts,
        out_path=tmpdir2.name,
        file_prefix="test_agg_gen_2",
    )

    _ngdf, _nvals = agg_gen.calculate_agg()

    outfile = Path(tmpdir.name) / "test_agg_gen_2.csv"

    assert outfile.exists()



"""Ancillary function to support core functions in helper.py."""
from __future__ import annotations

import json
import logging
import re
import sys
import time
import warnings
from collections.abc import Iterator
from datetime import datetime
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

import geopandas as gpd
import netCDF4
import numpy as np
import numpy.typing as npt
import pandas as pd
import shapely
import xarray as xr
from pyproj import CRS
from shapely.geometry import box
from shapely.geometry import LineString
from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon
from shapely.ops import split

from gdptools.data.odap_cat_data import CatGrids
from gdptools.data.odap_cat_data import CatParams


logger = logging.getLogger(__name__)


def _get_grid_cell_sindex(grid_cells):
    start = time.perf_counter()
    spatial_index = grid_cells.sindex
    # print(type(spatial_index))
    end = time.perf_counter()
    print(f"Spatial index generations finished in {round(end-start, 2)} second(s)")
    return spatial_index


def _reproject_for_weight_calc(poly, grid_cells, grid_out_crs, wght_gen_crs):
    start = time.perf_counter()
    # grid_cells.set_crs(grid_in_crs, inplace=True)
    grid_cells.to_crs(grid_out_crs, inplace=True)
    poly.to_crs(grid_out_crs, inplace=True)
    end = time.perf_counter()
    print(
        f"Reprojecting to epsg:{wght_gen_crs} finished in {round(end-start, 2)}"
        " second(s)"
    )
    return poly, grid_cells


def _check_grid_cell_crs(grid_cells):
    if not grid_cells.crs:
        error_string = f"grid_cells don't contain a valid crs: {grid_cells.crs}"
        raise ValueError(error_string)


def _check_feature_crs(poly):
    if not poly.crs:
        error_string = f"polygons don't contain a valid crs: {poly.crs}"
        raise ValueError(error_string)


def _check_poly_idx(poly, poly_idx):
    if poly_idx not in poly.columns:
        error_string = (
            f"Error: poly_idx ({poly_idx}) is not found in the poly ({poly.columns})"
        )
        raise ValueError(error_string)


def _get_print_on(numrows: int) -> int:
    """Return an interval to print progress of run_weights() function.

    Args:
        numrows (int): Number of rows: as in number of polygons

    Returns:
        int: Reasonable interval to print progress statements. Prints at about 10%
    """
    if numrows <= 10:  # pragma: no cover
        print_on = 1
    elif numrows <= 100:
        print_on = 10
    elif numrows <= 1000:
        print_on = 100
    elif numrows <= 10000:
        print_on = 1000
    elif numrows <= 100000:
        print_on = 10000
    else:
        print_on = 50000
    return int(print_on)


def _get_crs(crs_in: Any) -> CRS:
    """Return pyproj.CRS given integer or string.

    Args:
        crs_in (Any): integer: epsg code or pyproj string

    Returns:
        CRS: pyproj.CRS
    """
    # if type(crs_in) == int:
    #     in_crs = CRS.from_epsg(crs_in)
    # elif type(crs_in) == str:
    #     in_crs = CRS.from_proj4(crs_in)
    return CRS.from_user_input(crs_in)


def _get_cells_poly(  # noqa
    xr_a: Union[xr.Dataset, xr.DataArray],
    x: str,
    y: str,
    crs_in: Any,
    verbose: Optional[bool] = False,
) -> gpd.GeoDataFrame:
    """Get cell polygons associated with "nodes" in xarray gridded data.

    Args:
        xr_a (Union[xr.Dataset, xr.DataArray]): _description_
        x (str): _description_
        y (str): _description_
        crs_in (Any): _description_
        verbose (Optional[bool], optional): _description_. Defaults to False.

    Returns:
        gpd.GeoDataFrame: _description_
    """
    tlon = xr_a[x]
    tlat = xr_a[y]
    in_crs = _get_crs(crs_in)
    # if type(crs_in) == int:
    #     in_crs = CRS.from_epsg(crs_in)
    # elif type(crs_in) == str:
    #     in_crs = CRS.from_proj4(crs_in)

    # out_crs = _get_crs(4326)
    lon, lat = np.meshgrid(tlon, tlat)
    poly = []
    if verbose:
        logger.info("calculating surrounding cell vertices")
    start = time.perf_counter()
    lon_n = [
        lon[i, j]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]
    lon_jm1 = [
        lon[i, j - 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]
    lon_ip1_jm1 = [
        lon[i + 1, j - 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]
    lon_ip1 = [
        lon[i + 1, j]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]
    lon_ip1_jp1 = [
        lon[i + 1, j + 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]
    lon_jp1 = [
        lon[i, j + 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]
    lon_im1_jp1 = [
        lon[i - 1, j + 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]
    lon_im1 = [
        lon[i - 1, j]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]
    lon_im1_jm1 = [
        lon[i - 1, j - 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]

    lat_n = [
        lat[i, j]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]
    lat_jm1 = [
        lat[i, j - 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]
    lat_ip1_jm1 = [
        lat[i + 1, j - 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]
    lat_ip1 = [
        lat[i + 1, j]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]
    lat_ip1_jp1 = [
        lat[i + 1, j + 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]
    lat_jp1 = [
        lat[i, j + 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]
    lat_im1_jp1 = [
        lat[i - 1, j + 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]
    lat_im1 = [
        lat[i - 1, j]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]
    lat_im1_jm1 = [
        lat[i - 1, j - 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]
    end = time.perf_counter()
    if verbose:
        logger.info(
            f"finished calculating surrounding cell vertices in {round(end-start, 2)} second(s)"
        )

    # print(len(lon_n), len(lat_n), type(lon_n), np.shape(lon_n))
    numcells = len(lon_n)
    index = np.array(range(numcells))
    i_index = np.empty(numcells)
    j_index = np.empty(numcells)
    count = 0
    for i in range(1, lon.shape[0] - 1):
        for j in range(1, lon.shape[1] - 1):
            i_index[count] = i
            j_index[count] = j
            count += 1

    if verbose:
        logger.info("calculating cell 1 centroids")
    start = time.perf_counter()
    tpoly_1_lon = [
        [lon_n[i], lon_jm1[i], lon_ip1_jm1[i], lon_ip1[i]] for i in range(numcells)
    ]
    tpoly_1_lat = [
        [lat_n[i], lat_jm1[i], lat_ip1_jm1[i], lat_ip1[i]] for i in range(numcells)
    ]
    # print(len(tpoly_1_lon), tpoly_1_lon[0])
    newp = [
        Polygon(zip(tpoly_1_lon[i], tpoly_1_lat[i]))  # noqa B905
        for i in range(numcells)
    ]
    p1 = [p.centroid for p in newp]
    # print(type(newp), newp[0], len(p1))
    end = time.perf_counter()
    if verbose:
        logger.info(
            f"finished calculating cell 1 centroids in {round(end-start, 2)} second(s)"
        )
    del tpoly_1_lat, tpoly_1_lon

    if verbose:
        logger.info("calculating cell 2 centroids")
    start = time.perf_counter()
    tpoly_2_lon = [
        [lon_n[i], lon_ip1[i], lon_ip1_jp1[i], lon_jp1[i]] for i in range(numcells)
    ]
    tpoly_2_lat = [
        [lat_n[i], lat_ip1[i], lat_ip1_jp1[i], lat_jp1[i]] for i in range(numcells)
    ]
    # print(len(tpoly_2_lon), tpoly_2_lon[0])
    newp = [
        Polygon(zip(tpoly_2_lon[i], tpoly_2_lat[i]))  # noqa B905
        for i in range(numcells)
    ]
    p2 = [p.centroid for p in newp]
    end = time.perf_counter()
    if verbose:
        logger.info(
            f"finished calculating cell 2 centroids in {round(end-start, 2)} second(s)"
        )

    del tpoly_2_lat, tpoly_2_lon, newp

    if verbose:
        logger.info("calculating cell 3 centroids")
    start = time.perf_counter()
    tpoly_3_lon = [
        [lon_n[i], lon_jp1[i], lon_im1_jp1[i], lon_im1[i]] for i in range(numcells)
    ]
    tpoly_3_lat = [
        [lat_n[i], lat_jp1[i], lat_im1_jp1[i], lat_im1[i]] for i in range(numcells)
    ]
    # print(len(tpoly_3_lon), tpoly_3_lon[0])
    newp = [
        Polygon(zip(tpoly_3_lon[i], tpoly_3_lat[i]))  # noqa B905
        for i in range(numcells)
    ]
    p3 = [p.centroid for p in newp]
    end = time.perf_counter()
    if verbose:
        logger.info(
            f"finished calculating cell 3 centroids in {round(end-start, 2)} second(s)"
        )

    del tpoly_3_lat, tpoly_3_lon, newp

    if verbose:
        logger.info("calculating cell 4 centroids")
    start = time.perf_counter()
    tpoly_4_lon = [
        [lon_n[i], lon_im1[i], lon_im1_jm1[i], lon_jm1[i]] for i in range(numcells)
    ]
    tpoly_4_lat = [
        [lat_n[i], lat_im1[i], lat_im1_jm1[i], lat_jm1[i]] for i in range(numcells)
    ]
    # print(len(tpoly_3_lon), tpoly_3_lon[0])
    newp = [
        Polygon(zip(tpoly_4_lon[i], tpoly_4_lat[i]))  # noqa B905
        for i in range(numcells)
    ]
    p4 = [p.centroid for p in newp]
    end = time.perf_counter()
    if verbose:
        logger.info(
            f"finished calculating cell 4 centroids in {round(end-start, 2)} second(s)"
        )

    del tpoly_4_lat, tpoly_4_lon, newp
    del (
        lon_n,
        lon_jm1,
        lon_ip1_jm1,
        lon_ip1,
        lon_ip1_jp1,
        lon_jp1,
        lon_im1_jp1,
        lon_im1,
        lon_im1_jm1,
    )
    del (
        lat_n,
        lat_jm1,
        lat_ip1_jm1,
        lat_ip1,
        lat_ip1_jp1,
        lat_jp1,
        lat_im1_jp1,
        lat_im1,
        lat_im1_jm1,
    )

    if verbose:
        logger.info("creating bounding polygons")
    start = time.perf_counter()
    lon_point_list = [[p1[i].x, p2[i].x, p3[i].x, p4[i].x] for i in range(numcells)]
    lat_point_list = [[p1[i].y, p2[i].y, p3[i].y, p4[i].y] for i in range(numcells)]
    poly = [
        Polygon(zip(lon_point_list[i], lat_point_list[i]))  # noqa B905
        for i in range(numcells)
    ]
    end = time.perf_counter()
    if verbose:
        logger.info(
            f"finished creating bounding polygons in {round(end-start, 2)} second(s)"
        )

    if verbose:
        logger.info("reprojecting cells")
    start = time.perf_counter()
    # grd_shp = xr_a[var].values.shape
    df = pd.DataFrame({"i_index": i_index, "j_index": j_index})
    gmcells = gpd.GeoDataFrame(df, index=index, geometry=poly, crs=in_crs)
    # gmcells.to_crs(crs=out_crs, inplace=True)
    end = time.perf_counter()
    if verbose:
        logger.info(f"finished reprojecting cells in {round(end-start, 2)} second(s)")

    return gmcells


def _build_subset_cat(
    cat_params: CatParams,
    cat_grid: CatGrids,
    bounds: npt.NDArray[np.double],
    date_min: str,
    date_max: Optional[str] = None,
) -> Dict[Any, Any]:
    """Create a dictionary to use with xarray .sel() method to subset by time and space.

    Args:
        cat_params (CatParams): _description_
        cat_grid (CatGrids): _description_
        bounds (npt.NDArray[np.double]): _description_
        date_min (str): _description_
        date_max (str, optional): _description_. Defaults to None.

    Returns:
        dict: _description_
    """
    xname = cat_grid.X_name
    yname = cat_grid.Y_name
    # print(type(xname), type(yname))
    tname = cat_params.T_name
    minx = bounds[0]
    maxx = bounds[2]
    miny = bounds[1]
    maxy = bounds[3]
    gridorder = bool(cat_grid.toptobottom)
    if not gridorder:
        return (
            {
                xname: slice(minx, maxx),
                yname: slice(maxy, miny),
                tname: date_min,
            }
            if date_max is None
            else {
                xname: slice(minx, maxx),
                yname: slice(maxy, miny),
                tname: slice(date_min, date_max),
            }
        )

    elif date_max is None:
        return {
            xname: slice(minx, maxx),
            yname: slice(miny, maxy),
            tname: date_min,
        }

    else:
        return {
            xname: slice(minx, maxx),
            yname: slice(miny, maxy),
            tname: slice(date_min, date_max),
        }


def _read_shp_file(shp_file: Union[str, Path, gpd.GeoDataFrame]) -> gpd.GeoDataFrame:
    """Read shapefile.

    Args:
        shp_file (Union[str, gpd.GeoDataFrame]): _description_

    Returns:
        gpd.GeoDataFrame: _description_
    """
    if not isinstance(shp_file, gpd.GeoDataFrame):
        return gpd.read_file(shp_file)
    else:
        return shp_file


def _get_shp_file(
    shp_file: gpd.GeoDataFrame, cat_grid: CatGrids, is_degrees: bool
) -> Tuple[gpd.GeoDataFrame, npt.NDArray[np.double]]:
    """Return GeoDataFrame and bounds of shapefile.

    Args:
        shp_file (gpd.GeoDataFrame): _description_
        cat_grid (CatGrids): _description_
        is_degrees (bool): _description_

    Returns:
        Union[gpd.GeoDataFrame, npt.NDArray[np.double]y]: _description_
    """
    # read shapefile, calculate total_bounds, and project to grid's projection
    gdf = shp_file.to_crs(cat_grid.proj)
    # buffer polygons bounding box by twice max resolution of grid
    bbox = box(*gdf.total_bounds)
    gdf_bounds = bbox.buffer(2 * max(cat_grid.resX, cat_grid.resY)).bounds
    if is_degrees and (gdf_bounds[0] < -180.0) & (gdf_bounds[2] > 180.0):
        newxmax = 180 - (abs(gdf_bounds[0]) - 180.0)
        newxmin = -180.0 + (abs(gdf_bounds[2]) - 180.0)
        gdf_bounds = (newxmin, gdf_bounds[1], newxmax, gdf_bounds[3])
        print("Warning: The feature data crosses the anitmeridian.")

    # TODO: need to rethink buffer - leaving it out for now
    # gdf_bounds = gdf.total_bounds.buffer(2.0*buffer)
    return gdf, gdf_bounds


def _is_degrees(ds: xr.Dataset, cat_grid: CatGrids) -> bool:
    """Test if degrees in attributes on longitude.

    Args:
        ds (xr.Dataset): _description_
        cat_grid (CatGrids): _description_

    Returns:
        bool: _description_
    """
    xvals = ds[cat_grid.X_name]
    t_attr = xvals.attrs
    units = t_attr.get("units")
    return bool(re.search("degrees", units, flags=re.IGNORECASE))


def _is_degrees_nc(ds: xr.Dataset, x_name: str) -> bool:
    """Test if degrees in attributes on longitude.

    Args:
        ds (xr.Dataset): _description_
        x_name (str): _description_

    Returns:
        bool: _description_
    """
    xvals = ds[x_name]
    t_attr = xvals.attrs
    try:
        units = t_attr.get("units")
    except KeyError:
        tkey = "units"
        print(
            f" Error: key- {tkey} not found in coordinates of xarray"
            "dataset - make sure the coordinates of the dataset include"
            "a unit attribute per cf-conventions"
        )

    return bool(re.search("degrees", units, flags=re.IGNORECASE))


def _is_lon_0_360(vals: npt.NDArray[np.double]) -> bool:
    """Test if longitude spans 0-360.

    Args:
        vals (npt.NDArray[np.double]): _description_

    Returns:
        bool: _description_
    """
    result = False
    if (vals[0] > 180.0) & (np.min(vals) > 0.0):
        result = False
    elif (np.max(vals) > 180.0) & (np.min(vals) > 180.0):
        result = False
    elif np.max(vals) > 180.0:
        result = True

    return result


def _get_shp_bounds_w_buffer(
    gdf: gpd.GeoDataFrame, ds: xr.Dataset, crs: Any, lon: str, lat: str
) -> npt.NDArray[np.double]:
    """Return bounding box based on 2 * max(ds.dx, ds.dy).

    Args:
        gdf (gpd.GeoDataFrame): _description_
        ds (xr.Dataset): _description_
        crs (Any): _description_
        lon (str): _description_
        lat (str): _description_

    Returns:
        tuple: _description_
    """
    bbox = box(*gdf.to_crs(crs).total_bounds)
    return np.asarray(
        bbox.buffer(
            2 * max(max(np.diff(ds[lat].values)), max(np.diff(ds[lon].values)))
        ).bounds
    )


def _check_for_intersection(
    cat_params: CatParams,
    cat_grid: CatGrids,
    gdf: gpd.GeoDataFrame,
) -> Tuple[bool, bool, bool]:
    """Check broadly for intersection between features and grid.

    Args:
        cat_params (CatParams): _description_
        cat_grid (CatGrids): _description_
        gdf (gpd.GeoDataFrame): _description_

    Returns:
        Tuple[bool, bool, bool]: _description_
    """
    is_degrees = False
    is_intersect = True
    is_0_360 = False
    ds_url = cat_params.URL
    ds = xr.open_dataset(
        ds_url + "#fillmismatch", decode_coords=True, chunks={}
    )  # type:ignore
    xvals = ds[cat_grid.X_name]
    yvals = ds[cat_grid.Y_name]
    minx = xvals.values.min()
    maxx = xvals.values.max()
    miny = yvals.values.min()
    maxy = yvals.values.max()
    ds_bbox = box(minx, miny, maxx, maxy)
    bounds = _get_shp_bounds_w_buffer(
        gdf,
        ds,
        cat_grid.proj,
        cat_grid.X_name,
        cat_grid.Y_name,
    )
    is_degrees = _is_degrees(ds=ds, cat_grid=cat_grid)
    if is_degrees & (not ds_bbox.intersects(box(*np.asarray(bounds).tolist()))):
        is_intersect = False
        is_0_360 = _is_lon_0_360(xvals.values)
        if is_0_360:
            warning_string = (
                "0-360 longitude crossing the international date line encountered.\n"
                "Longitude coordinates will be 0-360 in output."
            )
            warnings.warn(warning_string)

    return is_intersect, is_degrees, is_0_360


def _check_for_intersection_nc(
    ds: xr.Dataset,
    x_name: str,
    y_name: str,
    proj: Any,
    gdf: gpd.GeoDataFrame,
) -> Tuple[bool, bool, bool]:
    """Check broadly for intersection between features and grid.

    Args:
        ds (xr.Dataset): _description_
        x_name (str): _description_
        y_name (str): _description_
        proj (Any): _description_
        gdf (gpd.GeoDataFrame): _description_

    Returns:
        Tuple[bool, bool, bool]: _description_
    """
    is_degrees = False
    is_intersect = True
    is_0_360 = False

    xvals = ds[x_name]
    yvals = ds[y_name]
    minx = xvals.values.min()
    maxx = xvals.values.max()
    miny = yvals.values.min()
    maxy = yvals.values.max()
    ds_bbox = box(minx, miny, maxx, maxy)
    bounds = _get_shp_bounds_w_buffer(
        gdf,
        ds,
        proj,
        x_name,
        y_name,
    )
    is_degrees = _is_degrees_nc(ds=ds, x_name=x_name)
    if is_degrees & (not ds_bbox.intersects(box(*np.asarray(bounds).tolist()))):
        is_intersect = False
        is_0_360 = _is_lon_0_360(xvals.values)
        if is_0_360:
            warning_string = (
                "0-360 longitude crossing the international date line encountered.\n"
                "Longitude coordinates will be 0-360 in output."
            )
            warnings.warn(warning_string)

    return is_intersect, is_degrees, is_0_360


def _get_data_via_catalog(
    cat_params: CatParams,
    cat_grid: CatGrids,
    bounds: npt.NDArray[np.double],
    begin_date: str,
    end_date: Optional[str] = None,
    rotate_lon: Optional[bool] = False,
) -> xr.DataArray:
    """Get xarray spatial and temporal subset.

    Args:
        cat_params (CatParams): _description_
        cat_grid (CatGrids): _description_
        bounds (np.ndarray): _description_
        begin_date (str): _description_
        end_date (Optional[str], optional): _description_. Defaults to None.
        rotate_lon (Optional[bool], optional): _description_. Defaults to False.

    Returns:
        xr.DataArray: _description_
    """
    ds_url = cat_params.URL
    ds = xr.open_dataset(
        ds_url + "#fillmismatch", decode_coords=True, chunks={}
    )  # type:ignore
    if rotate_lon:
        lon = cat_grid.X_name
        ds.coords[lon] = (ds.coords["lon"] + 180) % 360 - 180
        ds = ds.sortby(ds[lon])

    # get grid data subset to polygons buffered bounding box
    ss_dict = _build_subset_cat(cat_params, cat_grid, bounds, begin_date, end_date)
    # gridMET requires the '#fillmismatch' see:
    # https://discourse.oceanobservatories.org/
    # t/
    # accessing-data-on-thredds-opendap-via-python-netcdf4-or-xarray
    # -dealing-with-fillvalue-type-mismatch-error/61

    varname = cat_params.varname
    return ds[varname].sel(**ss_dict)  # type: ignore


def _get_wieght_df(wght_file: Union[str, pd.DataFrame], poly_idx: str) -> pd.DataFrame:
    if isinstance(wght_file, pd.DataFrame):
        # wghts = wght_file.copy()
        wghts = wght_file.astype({"i": int, "j": int, "wght": float, poly_idx: str})
    elif isinstance(wght_file, str):
        wghts = pd.read_csv(
            wght_file, dtype={"i": int, "j": int, "wght": float, poly_idx: str}
        )
    else:
        sys.exit("wght_file must be one of string or pandas.DataFrame")
    return wghts


def _date_range(p_start: str, p_end: str, intv: int) -> Iterator[str]:
    """Return a date range.

    Args:
        p_start (str): _description_
        p_end (str): _description_
        intv (int): _description_

    Yields:
        Iterator[str]: _description_
    """
    start = datetime.strptime(p_start, "%Y-%m-%d")
    end = datetime.strptime(p_end, "%Y-%m-%d")
    diff = (end - start) / intv
    for i in range(intv):
        yield (start + diff * i).strftime("%Y-%m-%d")
    yield end.strftime("%Y-%m-%d")


def _get_catalog_time_increment(param: dict) -> Tuple[int, str]:
    interval = param.get("interval").split(" ")
    return int(interval[0]), str(interval[1])


def _get_dataframe(object: Union[str, pd.DataFrame]) -> pd.DataFrame:
    if isinstance(object, str):
        return pd.DataFrame.from_dict(json.loads(object))
    else:
        return object


# this function is copied from here:
# https://github.com/gboeing/osmnx/blob/main/osmnx/utils_geo.py
# if it turns out to be helpful in speedin up intersections then I'll import it via
# the package - for now copied in here for development testing.
def _quadrat_cut_geometry(
    geometry: Union[shapely.geometry.Polygon, shapely.geometry.MultiPolygon],
    quadrat_width: float,
    min_num: Optional[int] = 3,
) -> shapely.geometry.MultiPolygon:
    """Split a Polygon or MultiPolygon up into sub-polygons of a specified size.

    Args:
        geometry (Union[shapely.geometry.Polygon, shapely.geometry.MultiPolygon]): _description_
        quadrat_width (float): _description_
        min_num (Optional[int], optional): _description_. Defaults to 3.

    Returns:
        shapely.geometry.MultiPolygon: _description_
    """
    # create n evenly spaced points between the min and max x and y bounds
    west, south, east, north = geometry.bounds
    x_num = int(np.ceil((east - west) / quadrat_width) + 1)
    y_num = int(np.ceil((north - south) / quadrat_width) + 1)
    x_points = np.linspace(west, east, num=max(x_num, min_num))
    y_points = np.linspace(south, north, num=max(y_num, min_num))

    # create a quadrat grid of lines at each of the evenly spaced points
    vertical_lines = [
        LineString([(x, y_points[0]), (x, y_points[-1])]) for x in x_points
    ]
    horizont_lines = [
        LineString([(x_points[0], y), (x_points[-1], y)]) for y in y_points
    ]
    lines = vertical_lines + horizont_lines

    # recursively split the geometry by each quadrat line
    geometries = [geometry]

    for line in lines:
        # split polygon by line if they intersect, otherwise just keep it
        split_geoms = [
            split(g, line).geoms if g.intersects(line) else [g] for g in geometries
        ]
        # now flatten the list and process these split geoms on the next line in
        # the list of lines
        geometries = [g for g_list in split_geoms for g in g_list]

    return MultiPolygon(geometries)


def _get_default_val(native_dtype: np.dtype):
    """Get default fill value based on gridded data dtype.

    Args:
        native_dtype (np.dtype): _description_

    Raises:
        TypeError: _description_

    Returns:
        _type_: _description_
    """
    if native_dtype.kind == "i":
        dfval = netCDF4.default_fillvals["i8"]
    elif native_dtype.kind == "f":
        dfval = netCDF4.default_fillvals["f8"]
    else:
        raise TypeError(
            "gdptools currently only supports int and float types."
            f"The value type here is {native_dtype}"
        )

    return dfval


def _get_interp_array(n_geo: int, nts: int, native_dtype: np.dtype, default_val: Any):
    """Get array for interpolation based on the dtype of the gridded data.

    Args:
        n_geo (int): _description_
        nts (int): _description_
        native_dtype (np.dtype): _description_
        default_val (Any): _description_

    Raises:
        TypeError: _description_

    Returns:
        _type_: _description_
    """
    if native_dtype.kind == "i":
        # val_interp = np.empty((nts, n_geo), dtype=np.dtype("int64"))
        val_interp = np.full(
            (nts, n_geo), dtype=np.dtype("int64"), fill_value=default_val
        )
    elif native_dtype.kind == "f":
        # val_interp = np.empty((nts, n_geo), dtype=np.dtype("float64"))
        val_interp = np.full(
            (nts, n_geo), dtype=np.dtype("float64"), fill_value=default_val
        )
    else:
        raise TypeError(
            "gdptools currently only supports int and float types."
            f"The value type here is {native_dtype}"
        )

    return val_interp


def _get_top_to_bottom(data: Union[xr.Dataset, xr.DataArray], y_coord: str) -> bool:
    """Get orientation of y-coordinate data in xarray Dataset.

    Args:
        data (xr.Dataset): _description_
        y_coord (str): _description_

    Returns:
        bool: _description_
    """
    yy = data.coords[y_coord].values
    ttb: bool = yy[0] <= yy[-1]
    return ttb



"""Placeholder __init__.py for data package."""
from .user_data import ODAPCatData
from .user_data import UserCatData
from .user_data import UserTiffData

__all__ = ["ODAPCatData", "UserCatData", "UserTiffData"]



"""Test cases for the __main__ module."""
import pytest
from click.testing import CliRunner

from gdptools import __main__


@pytest.fixture
def runner() -> CliRunner:
    """Fixture for invoking command-line interfaces."""
    return CliRunner()


def test_main_succeeds(runner: CliRunner) -> None:
    """It exits with a status code of zero."""
    result = runner.invoke(__main__.main)
    assert result.exit_code == 0



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['gdptools', 'gdptools.agg', 'gdptools.data', 'gdptools.weights', 'tests']

package_data = \
{'': ['*'],
 'tests': ['data/*',
           'data/DRB/*',
           'data/TM_WORLD_BORDERS_SIMPL-0.3/*',
           'data/rasters/TEXT_PRMS/*',
           'data/rasters/slope/*']}

install_requires = \
['Bottleneck>=1.3.3',
 'MetPy>=1.2.0',
 'Shapely<2.0.0',
 'attrs>=21.4.0',
 'dask>=2022.0.0',
 'geopandas>=0.11.0',
 'joblib>=1.1.0',
 'netCDF4<=1.6.1',
 'numpy<1.24.0',
 'pandas>=1.4.0',
 'pydantic>=1.9.0',
 'pygeos>=0.12.0',
 'pyproj>=3.3.0',
 'rasterio>=1.2.9',
 'rioxarray>=0.12.0',
 'scipy>=1.9.0',
 'xarray>=2022.6.0',
 'zarr>=2.12.0']

entry_points = \
{'console_scripts': ['gdptools = gdptools.__main__:main']}

setup_kwargs = {
    'name': 'gdptools',
    'version': '0.0.29',
    'description': 'Gdptools',
    'long_description': "# Readme\n\n[![PyPI](https://img.shields.io/pypi/v/gdptools.svg)](https://pypi.org/project/gdptools/)\n[![conda](https://anaconda.org/conda-forge/gdptools/badges/version.svg)](https://anaconda.org/conda-forge/gdptools)\n[![Latest Release](https://code.usgs.gov/wma/nhgf/toolsteam/gdptools/-/badges/release.svg)](https://code.usgs.gov/wma/nhgf/toolsteam/gdptools/-/releases)\n\n[![Status](https://img.shields.io/pypi/status/gdptools.svg)](https://pypi.org/project/gdptools/)\n[![Python Version](https://img.shields.io/pypi/pyversions/gdptools)](https://pypi.org/project/gdptools)\n\n[![License](https://img.shields.io/pypi/l/gdptools)](https://creativecommons.org/publicdomain/zero/1.0/legalcode)\n\n[![Read the documentation at https://gdptools.readthedocs.io/](https://img.shields.io/readthedocs/gdptools/latest.svg?label=Read%20the%20Docs)](https://gdptools.readthedocs.io/)\n[![pipeline status](https://code.usgs.gov/wma/nhgf/toolsteam/gdptools/badges/main/pipeline.svg)](https://code.usgs.gov/wma/nhgf/toolsteam/gdptools/-/commits/main)\n[![coverage report](https://code.usgs.gov/wma/nhgf/toolsteam/gdptools/badges/main/coverage.svg)](https://code.usgs.gov/wma/nhgf/toolsteam/gdptools/-/commits/main)\n\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://code.usgs.gov/pre-commit/pre-commit)\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://code.usgs.gov/psf/black)\n[![Poetry](https://img.shields.io/badge/poetry-enabled-blue)](https://python-poetry.org/)\n[![Conda](https://img.shields.io/badge/conda-enabled-green)](https://anaconda.org/)\n\n## Welcome\n\nWelcome to gdptools, a python package for grid- or polyon-to-polygon area-weighted interpolation statistics.\n\n![Welcome figure](./docs/assets/Welcom_fig.png)\n\n<figcaption>Example grid-to-polygon interpolation.  A) Huc12 basins for Delaware River Watershed. B) Gridded monthly water evaporation amount (mm) from TerraClimate dataset. C) Area-weighted-average interpolation of gridded TerraClimate data to Huc12 polygons.</figcaption>\n\n## Documentation\n\n[gdptools documentation](https://gdptools.readthedocs.io/en/latest/)\n\n## Features\n\n- Grid-to-polygon interpolation of area-weighted statistics.\n- Use [Mike Johnson's OPeNDAP catalog][1] to access over 1700 unique datasets.\n- Use any gridded dataset that can be read by xarray.\n- Uses spatial index methods for improving the efficiency of areal-wieght calculation detailed by [Geoff Boeing][2]\n\n[1]: https://mikejohnson51.github.io/opendap.catalog/articles/catalog.html\n[2]: https://geoffboeing.com/2016/10/r-tree-spatial-index-python/\n\n### Example catalog datasets\n\n| Dataset                                                                                    | Description                                                                                                                                                                                                                                                                                                    | Dates                      | Links |\n| ------------------------------------------------------------------------------------------ | -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- | -------------------------- | ----- |\n| [BCCA](https://gdo-dcp.ucllnl.org/downscaled_cmip_projections/dcpInterface.html#About)     | Bias Corrected Constructed Analogs V2 Daily Climate Projections (BACA) contains projections of daily BCCA CMIP3 and CMIP5 projections of precipitation, daily maximum, and daily minimum temperature over the contiguous United States                                                                         | 1950 - 2100                |       |\n| [BCSD](https://gdo-dcp.ucllnl.org/downscaled_cmip_projections/dcpInterface.html#About)     | Bias Corrected Spatially Downscaled (BCSD) Monthly CMIP5 Climate Projections                                                                                                                                                                                                                                   | 1950 - 2100                |       |\n| [CHIRPS](https://www.chc.ucsb.edu/data/chirps)                                             | Rainfall Estimates from Rain Gauge and Satellite Observations                                                                                                                                                                                                                                                  | 1980 - Current Month       |       |\n| [Daymet](https://daymet.ornl.gov/)                                                         | Daymet provides long-term, continuous, gridded estimates of daily weather and climatology variables by interpolating and extrapolating ground-based observations through statistical modeling techniques.                                                                                                      | 1980 through previous year |       |\n| [LOCA](https://gdo-dcp.ucllnl.org/downscaled_cmip_projections/dcpInterface.html#**About**) | LOCA, which stands for Localized Constructed Analogs, is a technique for downscaling climate model projections of the future climate.                                                                                                                                                                          | 1950 - 2100                |       |\n| [MACA](https://www.climatologylab.org/maca.html)                                           | Multivariate Adaptive Constructed Analogs (MACA) is a statistical method for downscaling Global Climate Models (GCMs) from their native coarse resolution to a higher spatial resolution that captures reflects observed patterns of daily near-surface meteorology and simulated changes in GCMs experiments. | 1950-2005 and 2006-2100    |       |\n| [PRISM-Monthly](https://cida.usgs.gov/thredds/catalog.html?dataset=cida.usgs.gov/prism_v2) | Parameter-elevation Regressions on Independent Slopes                                                                                                                                                                                                                                                          | 1895-2020                  |       |\n| [TerraClimate](https://www.climatologylab.org/terraclimate.html)                           | TerraClimate is a dataset of monthly climate and climatic water balance for global terrestrial surfaces from 1958-2019. These data provide important inputs for ecological and hydrological studies at global scales that require high spatial resolution and time-varying data.                               | 1958-2020                  |       |\n| [gridMET](https://www.climatologylab.org/gridmet.html)                                     | daily high-spatial resolution (~4-km, 1/24th degree) surface meteorological data covering the contiguous US                                                                                                                                                                                                    | 1979-yesterday             |       |\n\n## Data Requirements\n\n### Data - xarray (gridded data) and Geopandas (Polygon data)\n\n- [Xarray](https://docs.xarray.dev/en/stable/)\n\n  - Any endpoint that can be read by xarray and contains projected coordinates.\n    - The endpoint can be supplied by the OPeNDAP catalog or from a user-supplied end-point.\n  - Projection: any projection that can be read by proj.CRS (similar to Geopandas)\n\n- [Geopandas](https://geopandas.org/en/stable/)\n  - Any file that can be read by Geopandas\n  - Projection: any projection that can be read by proj.CRS\n\n## Installation\n\nYou can install _Gdptools_ via [pip](https://pip.pypa.io/) from [PyPI](https://pypi.org/):\n\n        pip install gdptools\n\nor install via [conda](https://anaconda.org/) from [conda-forge](https://anaconda.org/conda-forge/gdptools):\n\n       conda install -c conda-forge gdptools\n\n## Usage\n\nPlease see the example notebooks for detailes.\n\n### Catalog Examples\n\n- [OPeNDAP Catalog Example](./docs/terraclime_et.ipynb)\n\n### Non-catalog Examples\n\n- [Non-catalog example - gridMET](./docs/Gridmet_non_catalog.ipynb)\n- [Non-catalog example - Merra-2](./docs/Merra-2-example.ipynb)\n\n## Contributing\n\nContributions are very welcome. To learn more, see the Contributor Guide\\_.\n\n## License\n\nDistributed under the terms of the [CC0 1.0 Universal license](https://creativecommons.org/publicdomain/zero/1.0/legalcode), _Gdptools_ is free and open source software.\n\n## Issues\n\nIf you encounter any problems, please [file an issue](https://code.usgs.gov/wma/nhgf/toolsteam/gdptools/issues) along with a detailed description.\n\n## Credits\n\nThis project was generated from [@hillc-usgs](https://code.usgs.gov/hillc-usgs)'s [Pygeoapi Plugin Cookiecutter](https://code.usgs.gov/wma/nhgf/pygeoapi-plugin-cookiecutter) template.\n",
    'author': 'Richard McDonald',
    'author_email': 'rmcd@usgs.gov',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://code.usgs.gov/wma/nhgf/toolsteam/gdptools',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<=3.11',
}


setup(**setup_kwargs)



"""Calculate weights."""
import time
from typing import Any
from typing import Literal
from typing import Optional

import geopandas as gpd
import pandas as pd

from gdptools.data.user_data import UserData
from gdptools.data.weight_gen_data import WeightData
from gdptools.weights.calc_weight_engines import ParallelWghtGenEngine
from gdptools.weights.calc_weight_engines import SerialWghtGenEngine

WEIGHT_GEN_METHODS = Literal["serial", "parallel"]
""" Methods used in WghtGen class.

seerial: Iterates through polygons to calculate weights.  Sufficient for most cases.
parallel: Chunks polygons and distributes to available processors.  Provides a
    substantial speedup when there is a large number of polygons.

Raises:
    TypeError: If value is not one of "serial" or "parallel".

Returns:
    _type_: str
"""


class WeightGen:
    """Class for weight calculation."""

    def __init__(
        self,
        user_data: UserData,
        method: WEIGHT_GEN_METHODS,
        output_file: str,
        weight_gen_crs: Any,
        verbose: Optional[bool] = False,
    ) -> None:
        """Weight generation class.

        Args:
            user_data (UserData): _description_
            method (WEIGHT_GEN_METHODS): :data: `gdptools.weights.WeightGen.WEIGHT_GEN_METHODS`
            output_file (str): _description_
            weight_gen_crs (Any): _description_
            verbose (Optional[bool], optional): _description_. Defaults to False.

        Raises:
            TypeError: _description_
        """
        self._user_data = user_data
        self._method = method
        self._output_file = output_file
        self._weight_gen_crs = weight_gen_crs
        self._verbose = verbose
        self._weight_data = None
        self._intersections = None

        if self._method == "serial":
            self.__calc_method = SerialWghtGenEngine
            print("Using serial engine")
        elif self._method == "parallel":
            self.__calc_method = ParallelWghtGenEngine
            print("Using parallel engine")
        else:
            raise TypeError(f"method: {self._method} not in [serial, parallel]")

    def calculate_weights(self, intersections: Optional[bool] = False) -> pd.DataFrame:
        """Calculate weights.

        Args:
            intersections (Optional[bool], optional): _description_. Defaults to False.

        Returns:
            pd.DataFrame: _description_
        """
        tstrt = time.perf_counter()
        self._weight_data: WeightData = self._user_data.prep_wght_data()
        tend = time.perf_counter()
        print(f"Data preparation finished in {tend - tstrt:0.4f} seconds")
        if intersections:
            print("Saving interesections in weight generation.")
            weights, self._intersections = self.__calc_method().calc_weights(
                poly=self._weight_data.feature,
                poly_idx=self._weight_data.id_feature,
                grid_cells=self._weight_data.grid_cells,
                wght_gen_crs=self._weight_gen_crs,
                filename=self._output_file,
                intersections=intersections,
                verbose=self._verbose,
            )
        else:
            weights = self.__calc_method().calc_weights(
                poly=self._weight_data.feature,
                poly_idx=self._weight_data.id_feature,
                grid_cells=self._weight_data.grid_cells,
                wght_gen_crs=self._weight_gen_crs,
                filename=self._output_file,
                intersections=intersections,
                verbose=self._verbose,
            )
        return weights

    @property
    def grid_cells(self) -> gpd.GeoDataFrame:
        """Return grid_cells."""
        if self._weight_data.grid_cells is None:
            print("grid_cells not calculated yet. Run calculate_weights().")
        return self._weight_data.grid_cells

    @property
    def intersections(self) -> gpd.GeoDataFrame:
        """Return intersections."""
        if self._intersections is None:
            print(
                "intersections not calculated, "
                "Run calculate_weights(intersectiosn=True)"
            )
        return self._intersections



"""Primary functions for poly-to-poly area-weighted mapping."""
from __future__ import annotations

import itertools
import logging
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

import geopandas as gpd
import numpy as np
import numpy.typing as npt
import pandas as pd
import xarray as xr
from shapely.geometry import Polygon

from gdptools.data.odap_cat_data import CatGrids
from gdptools.data.odap_cat_data import CatParams
from gdptools.utils import _check_for_intersection
from gdptools.utils import _get_data_via_catalog
from gdptools.utils import _get_shp_file
from gdptools.utils import _read_shp_file

# from numba import jit

logger = logging.getLogger(__name__)

pd_offset_conv: Dict[str, str] = {
    "years": "Y",
    "months": "M",
    "days": "D",
    "hours": "H",
}


def get_cells_poly_2d(
    xr_a: xr.Dataset, lon_str: str, lat_str: str, in_crs: Any
) -> gpd.GeoDataFrame:
    """Get cell polygons associated with 2d lat/lon coordinates.

    Args:
        xr_a (xr.Dataset): _description_
        lon_str (str): _description_
        lat_str (str): _description_
        in_crs (Any): _description_

    Returns:
        gpd.GeoDataFrame: _description_
    """
    lon = xr_a[lon_str]
    lat = xr_a[lat_str]
    count = 0
    poly = []
    lon_n = [
        lon[i, j]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]
    lon_jm1 = [
        lon[i, j - 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]
    lon_ip1_jm1 = [
        lon[i + 1, j - 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]
    lon_ip1 = [
        lon[i + 1, j]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]
    lon_ip1_jp1 = [
        lon[i + 1, j + 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]
    lon_jp1 = [
        lon[i, j + 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]
    lon_im1_jp1 = [
        lon[i - 1, j + 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]
    lon_im1 = [
        lon[i - 1, j]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]
    lon_im1_jm1 = [
        lon[i - 1, j - 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lon.shape[1] - 1)
    ]

    lat_n = [
        lat[i, j]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]
    lat_jm1 = [
        lat[i, j - 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]
    lat_ip1_jm1 = [
        lat[i + 1, j - 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]
    lat_ip1 = [
        lat[i + 1, j]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]
    lat_ip1_jp1 = [
        lat[i + 1, j + 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]
    lat_jp1 = [
        lat[i, j + 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]
    lat_im1_jp1 = [
        lat[i - 1, j + 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]
    lat_im1 = [
        lat[i - 1, j]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]
    lat_im1_jm1 = [
        lat[i - 1, j - 1]
        for i in range(1, lon.shape[0] - 1)
        for j in range(1, lat.shape[1] - 1)
    ]

    # print(len(lon_n), len(lat_n), type(lon_n), np.shape(lon_n))
    numcells = len(lon_n)
    index = np.array(range(numcells))
    i_index = np.empty(numcells)
    j_index = np.empty(numcells)
    for count, (i, j) in enumerate(
        itertools.product(range(1, lon.shape[0] - 1), range(1, lon.shape[1] - 1))
    ):
        i_index[count] = i
        j_index[count] = j
    tpoly_1_lon = [
        [lon_n[i], lon_jm1[i], lon_ip1_jm1[i], lon_ip1[i]] for i in range(numcells)
    ]
    tpoly_1_lat = [
        [lat_n[i], lat_jm1[i], lat_ip1_jm1[i], lat_ip1[i]] for i in range(numcells)
    ]
    # print(len(tpoly_1_lon), tpoly_1_lon[0])
    newp = [
        Polygon(zip(tpoly_1_lon[i], tpoly_1_lat[i]))  # noqa B905
        for i in range(numcells)
    ]
    p1 = [p.centroid for p in newp]
    # print(type(newp), newp[0], len(p1))

    tpoly_2_lon = [
        [lon_n[i], lon_ip1[i], lon_ip1_jp1[i], lon_jp1[i]] for i in range(numcells)
    ]
    tpoly_2_lat = [
        [lat_n[i], lat_ip1[i], lat_ip1_jp1[i], lat_jp1[i]] for i in range(numcells)
    ]
    print(len(tpoly_2_lon), tpoly_2_lon[0])
    newp = [
        Polygon(zip(tpoly_2_lon[i], tpoly_2_lat[i]))  # noqa B905
        for i in range(numcells)
    ]
    p2 = [p.centroid for p in newp]

    tpoly_3_lon = [
        [lon_n[i], lon_jp1[i], lon_im1_jp1[i], lon_im1[i]] for i in range(numcells)
    ]
    tpoly_3_lat = [
        [lat_n[i], lat_jp1[i], lat_im1_jp1[i], lat_im1[i]] for i in range(numcells)
    ]
    # print(len(tpoly_3_lon), tpoly_3_lon[0])
    newp = [
        Polygon(zip(tpoly_3_lon[i], tpoly_3_lat[i]))  # noqa B905
        for i in range(numcells)
    ]
    p3 = [p.centroid for p in newp]

    tpoly_4_lon = [
        [lon_n[i], lon_im1[i], lon_im1_jm1[i], lon_jm1[i]] for i in range(numcells)
    ]
    tpoly_4_lat = [
        [lat_n[i], lat_im1[i], lat_im1_jm1[i], lat_jm1[i]] for i in range(numcells)
    ]
    # print(len(tpoly_3_lon), tpoly_3_lon[0])
    newp = [
        Polygon(zip(tpoly_4_lon[i], tpoly_4_lat[i]))  # noqa B905
        for i in range(numcells)
    ]
    p4 = [p.centroid for p in newp]

    lon_point_list = [[p1[i].x, p2[i].x, p3[i].x, p4[i].x] for i in range(numcells)]
    lat_point_list = [[p1[i].y, p2[i].y, p3[i].y, p4[i].y] for i in range(numcells)]

    poly = [
        Polygon(zip(lon_point_list[i], lat_point_list[i]))  # noqa B905
        for i in range(numcells)
    ]

    df = pd.DataFrame({"i_index": i_index, "j_index": j_index})
    # tpoly_1 = [Polygon(x) for x in newp]
    # p1 = tpoly_1.centroid
    return gpd.GeoDataFrame(df, index=index, geometry=poly, crs=in_crs)


def build_subset(
    bounds: npt.NDArray[np.double],
    xname: str,
    yname: str,
    tname: str,
    toptobottom: bool,
    date_min: str,
    date_max: Optional[str] = None,
) -> Dict[str, object]:
    """Create a dictionary to use with xarray .sel() method to subset by time and space.

    Args:
        bounds (np.ndarray): _description_
        xname (str): _description_
        yname (str): _description_
        tname (str): _description_
        toptobottom (bool): _description_
        date_min (str): _description_
        date_max (Optional[str], optional): _description_. Defaults to None.

    Returns:
        dict: _description_
    """
    minx = bounds[0]
    maxx = bounds[2]
    miny = bounds[1]
    maxy = bounds[3]
    if not toptobottom:
        return (
            {
                xname: slice(minx, maxx),
                yname: slice(maxy, miny),
                tname: date_min,
            }
            if date_max is None
            else {
                xname: slice(minx, maxx),
                yname: slice(maxy, miny),
                tname: slice(date_min, date_max),
            }
        )

    elif date_max is None:
        return {
            xname: slice(minx, maxx),
            yname: slice(miny, maxy),
            tname: date_min,
        }

    else:
        return {
            xname: slice(minx, maxx),
            yname: slice(miny, maxy),
            tname: slice(date_min, date_max),
        }


def build_subset_tiff(
    bounds: npt.NDArray[np.double],
    xname: str,
    yname: str,
    toptobottom: bool,
    bname: str,
    band: int,
) -> Dict[str, object]:
    """Create a dictionary to use with xarray .sel() method to subset by time and space.

    Args:
        bounds (npt.NDArray[np.double]): _description_
        xname (str): _description_
        yname (str): _description_
        toptobottom (bool): _description_
        bname (str): _description_
        band (int): _description_

    Returns:
        Dict[str, object]: _description_
    """
    minx = bounds[0]
    maxx = bounds[2]
    miny = bounds[1]
    maxy = bounds[3]
    ss_dict = {}
    ss_dict = (
        {
            xname: slice(minx, maxx),
            yname: slice(miny, maxy),
            bname: band,
        }
        if toptobottom
        else {
            xname: slice(minx, maxx),
            yname: slice(maxy, miny),
            bname: band,
        }
    )

    return ss_dict


def build_subset_tiff_da(
    bounds: npt.NDArray[np.double],
    xname: str,
    yname: str,
    toptobottom: bool,
) -> Dict[str, object]:
    """Create a dictionary to use with xarray .sel() method to subset by time and space.

    Args:
        bounds (npt.NDArray[np.double]): _description_
        xname (str): _description_
        yname (str): _description_
        toptobottom (bool): _description_

    Returns:
        Dict[str, object]: _description_
    """
    minx = bounds[0]
    maxx = bounds[2]
    miny = bounds[1]
    maxy = bounds[3]
    ss_dict = {}
    ss_dict = (
        {
            xname: slice(minx, maxx),
            yname: slice(miny, maxy),
        }
        if toptobottom
        else {
            xname: slice(minx, maxx),
            yname: slice(maxy, miny),
        }
    )

    return ss_dict


def get_data_subset_catalog(
    cat_params: CatParams,
    cat_grid: CatGrids,
    shp_file: Union[str, gpd.GeoDataFrame],
    begin_date: str,
    end_date: str,
) -> xr.DataArray:
    """Get xarray subset data.

    Args:
        cat_params (CatParams): _description_
        cat_grid (CatGrids): _description_
        shp_file (Union[str, gpd.GeoDataFrame]): _description_
        begin_date (str): _description_
        end_date (str): _description_

    Returns:
        xr.DataArray: _description_
    """
    # run check on intersection of shape features and gridded data
    gdf = _read_shp_file(shp_file)
    is_intersect, is_degrees, is_0_360 = _check_for_intersection(
        cat_params=cat_params, cat_grid=cat_grid, gdf=gdf
    )

    # read shapefile, calculate total_bounds, and project to grid's projection
    gdf, gdf_bounds = _get_shp_file(
        shp_file=gdf, cat_grid=cat_grid, is_degrees=is_degrees
    )

    rotate_ds = bool((not is_intersect) & is_degrees & (not is_0_360))
    return _get_data_via_catalog(
        cat_params=cat_params,
        cat_grid=cat_grid,
        bounds=gdf_bounds,
        begin_date=begin_date,
        end_date=end_date,
        rotate_lon=rotate_ds,
    )



"""Tests for raster functions."""
from pathlib import Path
from tempfile import TemporaryDirectory

import geopandas as gpd
import pandas as pd
import pytest
import rioxarray as rxr
import xarray as xr

from gdptools import ZonalGen
from gdptools.data.user_data import UserTiffData


@pytest.fixture(scope="module")
def get_tiff_slope() -> xr.DataArray:
    """Get tiff slope file."""
    return rxr.open_rasterio("./tests/data/rasters/slope/slope.tif")


@pytest.fixture(scope="module")
def get_tiff_text() -> xr.DataArray:
    """Get tiff text_prms file."""
    return rxr.open_rasterio("./tests/data/rasters/TEXT_PRMS/TEXT_PRMS.tif")


@pytest.fixture(scope="module")
def get_gdf() -> gpd.GeoDataFrame:
    """Get gdf file."""
    return gpd.read_file("./tests/data/Oahu.shp")


@pytest.mark.parametrize(
    "vn,xn,yn,bd,bn,crs,cat,ds,gdf,fid",
    [
        (
            "slope",
            "x",
            "y",
            1,
            "band",
            26904,
            False,
            "get_tiff_slope",
            "get_gdf",
            "fid",
        ),
        (
            "TEXT_PRMS",
            "x",
            "y",
            1,
            "band",
            26904,
            True,
            "get_tiff_text",
            "get_gdf",
            "fid",
        ),
    ],
)
def test_cat_tiff_intersection(vn, xn, yn, bd, bn, crs, cat, ds, gdf, fid, request):
    """Test tiff intersection function."""
    data = UserTiffData(
        var=vn,
        ds=request.getfixturevalue(ds),
        proj_ds=crs,
        x_coord=xn,
        y_coord=yn,
        bname=bn,
        band=bd,
        f_feature=request.getfixturevalue(gdf),
        id_feature=fid,
        proj_feature=crs,
    )
    tmpdir = TemporaryDirectory()
    zonal_gen = ZonalGen(
        user_data=data,
        zonal_engine="serial",
        zonal_writer="csv",
        out_path=tmpdir.name,
        file_prefix="tmpzonal",
    )
    stats = zonal_gen.calculate_zonal(categorical=cat)

    assert isinstance(stats, pd.DataFrame)
    file = Path(tmpdir.name) / "tmpzonal.csv"
    assert file.exists()
    # tdata = TiffAttributes(
    #     varname=vn,
    #     xname=xn,
    #     yname=yn,
    #     band=bd,
    #     bname=bn,
    #     crs=crs,
    #     categorical=cat,
    #     ds=request.getfixturevalue(ds),
    # )

    # ngdf = tiff_intersection_by_feature(
    #     data=tdata, gdf=request.getfixturevalue(gdf), poly_idx=fid
    # )

    # assert isinstance(ngdf, pd.DataFrame)



"""Command line script for filling missing data in file."""
import sys
from pathlib import Path
from typing import Optional

import geopandas as gpd
import pandas as pd
import xarray

from gdptools.utils import _get_default_val


def fill_onhm_ncf(  # noqa:C901
    nfile: str,
    output_dir: str,
    var: str,
    lat: str,
    lon: str,
    feature_id: str,
    mfile: str,
    genmap: Optional[bool] = False,
) -> None:
    """Function uses nearest-neighbor, to fill missing feature values.

    Args:
        nfile (str): _description_
        output_dir (str): _description_
        var (str): _description_
        lat (str): _description_
        lon (str): _description_
        feature_id (str): _description_
        mfile (str): _description_
        genmap (Optional[bool], optional): _description_. Defaults to False.
    """
    odir = Path(output_dir)
    if not odir.exists():
        print(f"Path: {odir} does not exist")
        exit
    data = xarray.open_dataset(nfile, engine="netcdf4")  # type: ignore

    if var not in list(data.keys()):
        print(f"Error: {var} not in dataset")
        exit
    try:
        dfval = _get_default_val(data.var.dtype)
    except TypeError as e:
        print(e)
    if genmap:
        # create geodatafrom from masked data of missing values
        data_1d = data.isel(time=[0])
        df_mask = data_1d.where(data_1d >= dfval, drop=True)
        if df_mask.dims.get("nhru") == 0:
            sys.exit("No missing data - exiting")
        tmax_m_vals = df_mask[var].values[0, :]
        lon_m = df_mask[lon].values[:]
        lat_m = df_mask[lat].values[:]
        hruid_m = df_mask[feature_id].values
        df_m = pd.DataFrame(
            {"featid": hruid_m, "lon": lon_m, "lat": lat_m, "tmax": tmax_m_vals}
        )
        gdf_m = gpd.GeoDataFrame(df_m, geometry=gpd.points_from_xy(df_m.lon, df_m.lat))

        # create geodatafrom from non-missing data
        df_filled = data_1d.where(data.tmax < dfval, drop=True)
        tmax_f_vals = df_filled[var].values[0, :]
        lon_f = df_filled[lon].values[:]
        lat_f = df_filled[lat].values[:]
        hruid_f = df_filled[feature_id].values
        df_f = pd.DataFrame(
            {"featid": hruid_f, "lon": lon_f, "lat": lat_f, "tmax": tmax_f_vals}
        )
        gdf_f = gpd.GeoDataFrame(df_f, geometry=gpd.points_from_xy(df_f.lon, df_f.lat))

        # use spatial-join to find nearest filled data for each missing hru-id
        nearest_m = gdf_m.sjoin_nearest(gdf_f, distance_col="distance")

        # print(nearest_m.head())
        nearest_m.drop(
            ["lon_left", "lat_left", "lon_right", "lat_right"], axis=1
        ).to_csv(odir / "fill_missing_nearest.csv")
    else:
        nearest_m = pd.read_csv(mfile)

    miss_index = nearest_m.featid_left.values
    fill_index = nearest_m.featid_right.values

    # fill missing values
    for dvar in data.data_vars:
        if dvar != "crs":
            print(f"processing {dvar}")
            data["tmax"].loc[dict(nhru=miss_index)] = (
                data["tmax"].loc[dict(nhru=fill_index)].values
            )

    oldfile = Path(nfile)
    newfile = oldfile.parent / f"{oldfile.name[:-3]}_filled.nc"

    # write new netcdf file with _filled appended to existing filename
    encoding = {}
    encoding_keys = "_FillValue"
    for data_var in data.data_vars:
        encoding[data_var] = {
            key: value
            for key, value in data[data_var].encoding.items()
            if key in encoding_keys
        }
        encoding[data_var].update(zlib=True, complevel=2)

    for data_var in data.coords:
        encoding[data_var] = {
            key: value
            for key, value in data[data_var].encoding.items()
            if key in encoding_keys
        }
        encoding[data_var].update(_FillValue=None, zlib=True, complevel=2)
    print(encoding)
    data.to_netcdf(path=newfile, encoding=encoding)
